#!/usr/bin/perl
#
# Fix the GnuPlot generated EPS bounding box size.
#
# Process all EPS files found in the given directory, and its
# subdirectories (if any).

use Cwd;
use File::Find;
use File::Copy;

use strict;
package main;

$main::program = 'change-eps-colors';

initialize ();
run ();
terminate ();

#-------------------------------------------------------
sub run
{
   initialize_colors ();
   find_files ();
   fix_files ();
}

#-------------------------------------------------------
sub initialize_colors
{
   @main::colors = split /,/, $main::ini{'NewColors9'};
}

#-------------------------------------------------------
sub fix_files
{
   foreach (@main::files) {
      my ($fileName) = $_;
      fix_bounding_box ($fileName);
      change_colors ($fileName);
   }
}

#-------------------------------------------------------
sub change_colors
{
   my ($epsFile) = @_;

   open (EPS, $epsFile) or die "Could not open [$epsFile]\n";
   my (@eps) = <EPS>;
   close (EPS);

   my ($line);
   line:
   foreach $line (@eps) {
      # line types:
      if ($line =~ m|^(/LT)(\d)( .+?] )(.+?)( DL } def)$|) {
         my $s1 = $1;
         my $s2 = $2;
         my $s3 = $3;
         my $s4 = $4;
         my $s5 = $5;

         my $newColor = $main::colors[$s2];
         $line = "$s1$s2$s3$newColor$s5\n";
      }
   }

   open (EPS, ">$epsFile") or die "Could not open [$epsFile]\n";
   print EPS @eps;
   close (EPS);  
}

#-------------------------------------------------------
sub fix_bounding_box
{
   my ($epsFile) = @_;

   open (EPS, $epsFile) or die "Could not open [$epsFile]\n";
   my (@eps) = <EPS>;
   close (EPS);

   my ($line);
   line:
   foreach $line (@eps) {
      if ($line =~ m|(%%BoundingBox: \d+ \d+) (\d+) (\d+)|) {
         my ($s1) = $1;
         my ($s2) = $2;
         my ($s3) = $3;
         my ($boxWidth) = $main::ini{'EpsBoxAspectRatio'} * $s3;
         $boxWidth = sprintf ("%d", $boxWidth);
         $line = "$s1 $boxWidth $s3\n";
         last line;
      }
   }

   open (EPS, ">$epsFile") or die "Could not open [$epsFile]\n";
   print EPS @eps;
   close (EPS);  
}


#=======================================================

#-------------------------------------------------------
sub initialize
{
	my ($line, $inifile, $logfile);

	if (@ARGV) {
		$inifile = $ARGV[0];
	}
	else {
		$inifile = "$main::program.ini";
	}
	read_inifile ($inifile);

	$logfile = "$main::program.log";
	open (LOG, ">$logfile") or die $!
}

#-------------------------------------------------------
sub read_inifile
{
	my ($inifile) = @_;
	my ($line, $lineRaw, $key, $value, $contin);

	open (INI, $inifile) or die "could not open initialization file [$inifile]!\n";

	$contin = 0;
	$line = "";

	lineRaw:
	foreach $lineRaw (<INI>) {
		chomp ($lineRaw);
		
		# Allow for continuation lines:

		if ($lineRaw =~ m|\\$|) {
			$contin = 1;
		   $lineRaw =~ s|\\$||;
			$line .= $lineRaw;
			next lineRaw;
		}
		else {
			if ($contin) {
				$line .= $lineRaw;
			}
			else {
				$line = $lineRaw;
			}
			$contin = 0;
		}

		$line =~ s|\;.+||g;
		if (!($line =~ m|=|)) {
			next;
		}

		($key, $value) = split /=/, $line;
		$key =~ s|^\s+||;
		$key =~ s|\s+$||;
		$value =~ s|^\s+||;
		$value =~ s|\s+$||;
		$main::ini{$key} = $value;

		$line = "";
	}
	close (INI);
}

#-------------------------------------------------
sub find_files_dir_ext
{
	my ($directory, $extension) = @_;

	undef @main::files;

	$main::match = "";
	if (length($extension) > 0) {
		$main::match = "\.$extension";
	}
	find (\&wanted_files, $directory);

	return @main::files;
}

#-------------------------------------------------
sub find_all_files
{
	my (@tmp);

	@tmp = find_file_extension ("$main::ini{'InputDirectory'}", "", "");

	return @tmp;
}

#-------------------------------------------------
sub find_files
{
	my (@tmp);

	@tmp = find_file_extension ("$main::ini{'InputDirectory'}", "", $main::ini{'FileExtension'});

	return @tmp;
}

#-------------------------------------------------
sub find_file_extension
{
	my ($directory, $file, $extension) = @_;

	undef @main::files;

	$main::match = "";
	if (defined($file) && length($file) > 0) {
		$main::match = $file;
	}
	elsif (length($extension) > 0) {
		$main::match = "\.$extension";
	}

	find (\&wanted_files, $directory);

	return @main::files;
}

#-------------------------------------------------
sub wanted_files
{
	m|$main::match$|i && push @main::files, $File::Find::name;
}

#-------------------------------------------------------
sub terminate
{
	close (LOG);
	print "All done.  Press ENTER to terminate.";
	<STDIN>;
}
