#!/usr/bin/perl

use Cwd;
use File::Find;
use File::Copy;
use File::Path qw(make_path);

use strict;
package main;

($main::program = $0) =~ s,.*[\\/],,;
$main::program =~ s|\.pl||;
$main::program =~ s|\.exe||;

$| = 1;  # turn off buffering of STDOUT output

initialize ();

my $inputDir = "$main::ini{'DynaFitDirectory'}/$main::ini{'ProjectDirectory'}";
$inputDir =~ s|\.\.\.|\.\.|;
$main::ini{'InputDirectory'} = $inputDir;

find_files ();
filter_files ();
create_report_files ();
run ();
close_report_files ();
terminate ();

#-------------------------------------------------------
sub run
{
   print <<_END;

============================================================
$main::ini{'Title'}
============================================================

_END

   read_script_template ();
   process_files ();
   analyze_acceptance ();
   rewrite_data_sheets ();
   write_report ();
}

#-------------------------------------------------------
sub create_report_files
{
   my $path = $main::ini{'HtmlHeaderFile'};
   open (IN, $path) or error_exit ("could not open file\n[$path]!\n");
   @main::htmlHeader = <IN>;
   close (IN);

   $path = $main::ini{'HtmlFooterFile'};
   open (IN, $path) or error_exit ("could not open file\n[$path]!\n");
   @main::htmlFooter = <IN>;
   close (IN);

   my $status = mkdir ($main::ini{'ReportDirectory'});
   $path = "$main::ini{'ReportDirectory'}/$main::ini{'ReportFile'}.txt";
   open (TXT, ">$path") or error_exit ("could not open output file\n[$path]\n");
   
   $path = "$main::ini{'ReportDirectory'}/$main::ini{'ReportFile'}.html";
   open (HTM, ">$path") or error_exit ("could not open output file\n[$path]\n");
   
   print HTM @main::htmlHeader; 
}

#-------------------------------------------------------
sub close_report_files
{
   print HTM @main::htmlFooter; 

   close (TXT);
   close (HTM);

   my $path = "$main::ini{'ReportDirectory'}/$main::ini{'ReportFile'}.html";
   my $command = "start $path";
   system ($command);
}

#-------------------------------------------------------
sub filter_files
{
   my @files;
   foreach (@main::files) {
      my $path = $_;
      if ($path =~ m|$main::ini{'SheetFileName'}\.$main::ini{'FileExtension'}|) {
         push @files, $path;
      }
   }
   @main::files = @files;
}

#-------------------------------------------------------
sub rewrite_data_sheets
{
   foreach (@main::files) {
      my $path = $_;
      my $pathSave = $path;
      $path =~ s|$main::ini{'InputDirectory'}/||;

      my @fields = split /\//, $path;     
      if ($fields[-1] eq "sheet.txt") { pop @fields };
      if ($fields[-1] eq "data") { pop @fields };
      my ($replicate) = pop @fields;
      my ($compound) = pop @fields;
      
      my $record;
      if ($main::ini{'UseReplicates'}) {
         $record = $main::recordAcceptCompound{$compound};
      }
      else {
         my $key = "$compound/$replicate";
         $record = $main::recordAccept{$key};
      }
      
      open (DAT, $pathSave) or error_exit ("could not open file\n[$pathSave]\n");
      my @txt = <DAT>;
      close (DAT);
      
      my @txtOut;
      foreach (@txt) {
         my $line = $_;
         if ($line =~ m|^INCLUDE|) {
            push @txtOut, "INCLUDE\t$record\n";            
         }
         else {
            push @txtOut, $line;            
         }
      }
      print "rewriting [$pathSave]...\n";     
      open (DAT, ">$pathSave") or error_exit ("could not open output file\n[$pathSave]\n");
      print DAT @txtOut;
      close (DAT);
   }   
}

#-------------------------------------------------------
sub analyze_acceptance
{
   my @acceptColumn;
   foreach (sort keys %main::recordAccept) {
      my $key = $_;
      my @fields = split /\//, $key;
      my ($replicate) = pop @fields;
      my ($compound) = pop @fields;
      $key = "$compound/$replicate";
      $replicate =~ s|R||;
      if ($replicate == 1) {
         undef @acceptColumn;
         my $i = 0;
         for (; $i < $main::ini{'Columns'}; $i++) {
            push @acceptColumn, "+";
         }
      }
      my @record = split /\t/, $main::recordAccept{$key};
      my $i = 0;
      for (; $i < $main::ini{'Columns'}; $i++) {
         if ($record[$i] ne "+") {
            $acceptColumn[$i] = "-";
         }
      }
      my $record = join "\t", @acceptColumn;
      if ($main::ini{'UseReplicates'}) {
         if ($replicate == $main::ini{'Replicates'}) {
            $main::recordAcceptCompound{$compound} = $record;
         }
      }
      else {        
         $main::recordAcceptCompound{$compound} = $record;
      }
   }
}

#-------------------------------------------------------
sub write_report
{
   write_report_txt ();
   write_report_html ();
}

#-------------------------------------------------------
sub write_report_txt
{
   print TXT "Select raw data for kinetic analysis (\"X\" = excluded well)\n\n";
   print TXT "COMPOUND";
   my $n = $main::ini{'Columns'};
   my $i = 1;
   for ( ; $i <= $n; ++$i) {
      my $label = sprintf ("%s%2.2d", "col-", $i);
      print TXT "\t$label";
   }
   print TXT "\n";
   foreach (sort keys %main::recordAcceptCompound) {
      my $key = $_;
      my $record = $main::recordAcceptCompound{$key};
      $record =~ s|\+|.|g;
      $record =~ s|\-|X|g;
      print TXT "$key\t$record\n";
   }
   print TXT "\n\n";
}

#-------------------------------------------------------
sub write_report_html
{
   my $n = $main::ini{'Columns'};
   print HTM <<_END;  
<h2>$main::ini{'Title'}</h2>
_END

   if ($main::ini{'UseReplicates'}) {
      print HTM <<_END;  
<h3>Acceptance per compound</h3>

<table>
<tr><td class=\"th\">COMPOUND
_END
         my $i = 1;
         for ( ; $i <= $n; ++$i) {
            my $label = sprintf ("%s%2.2d", "col-", $i);
            print HTM "<td class=\"th\">\t$label";
         }
         print HTM "\n";
         foreach (sort keys %main::recordAcceptCompound) {
            my $key = $_;
            my $record = $main::recordAcceptCompound{$key};
            $record =~ s|\+|.|g;
            $record =~ s|\-|X|g;
            $record =~ s|\t|<td>\t|g;
            print HTM "<tr><td>$key<td>\t$record\n";
         }
         print HTM <<_END;
</table>
_END
   }
   print HTM <<_END;
<h3>Acceptance per replicate</h3>

<table>
<tr><td class=\"th\">CPD/REPL
_END
   my $i = 1;
   for ( ; $i <= $n; ++$i) {
      my $label = sprintf ("%s%2.2d", "col-", $i);
      print HTM "<td class=\"th\">\t$label";
   }
   print HTM "\n";
   foreach (sort keys %main::recordAccept) {
      my $key = $_;
      my $link = create_link ($key);
      my $record = $main::recordAccept{$key};
      $record =~ s|\+|.|g;
      $record =~ s|\-|X|g;
      $record =~ s|\t|<td valign=\"top\">\t|g;
      
      if ($main::ini{'UseReplicates'}) {
         my $subdir = sprintf ("%s%d", "R", $main::ini{'Replicates'});     
         if ($link =~ m|/$subdir|) {
            $link =~ s|/a>|/a><hr noshade>|;
         }
      }
      print HTM "<tr><td valign=\"top\">$link<td valign=\"top\">\t$record\n";
   }
   print HTM <<_END;
</table>
_END
}

#-------------------------------------------------------
sub create_link
{
   my ($key) = @_;
   my ($compound, $replicate) = split /\//, $key;
   my $prefix = "../";  # we have to back-track from the "results" directory!   This should be made programmatic.
   
   my $link = "<a href=\"$prefix$main::ini{'InputDirectory'}/$compound/$replicate/output/$main::ini{'ScriptFile'}/index.html\">";
   $link =~ s|\.txt||;
   $link .= "$key</a>";
   return $link;
}

#-------------------------------------------------------
sub read_script_template
{
   my $path = $main::ini{'ScriptTemplate'};
   open (IN, $path) or error_exit ("could not open file\n[$path]!\n");
   @main::scriptTemplate = <IN>;
   close (IN);
}

#-------------------------------------------------------
sub process_files
{
   foreach (@main::files) {
      my $path = $_;
      process_script ($path);
   }
}

#-------------------------------------------------------
sub process_script
{
   my ($path) = @_;
   my $script = write_script ($path);
   execute_script ($script);
   parse_output ($script);
}

#-------------------------------------------------------
sub parse_output
{
   my ($script) = @_;
   my $outputDir = $script;
   $outputDir =~ s|$main::ini{'ScriptFile'}|output|;
   my $outputSubdir = $main::ini{'ScriptFile'};
   $outputSubdir =~ s|\.txt||;

   $script =~ s|$main::ini{'InputDirectory'}/||;
   $script =~ s|/$main::ini{'ScriptFile'}||;

   my @fields = split /\//, $script;
   my ($replicate) = pop @fields;
   my ($compound) = pop @fields;
   $script = "$compound/$replicate";
   
   my $record = "$script";
   my $label = $record;
   my $recordAccept = "";
   my $n = $main::ini{'Columns'};
   my $i;
   for ($i = 0; $i < $n; ++$i) {
      my $iTask = $i + 1;
      my $sTask = sprintf ("%3.3d", $iTask);
      my $paramFile = "$outputDir/$outputSubdir/html/parameters-t" . $sTask . ".html";

      open (IN, $paramFile) or error_exit ("could not open file\n[$paramFile]\n");
      my @txt = <IN>;
      close (IN);
      
      $script =~ s|/$main::ini{'ScriptFile'}||;
      $script =~ s|$main::ini{'InputDirectory'}/||;
      $script =~ s|/R|\tR|;
      my @triplet;
      foreach (@txt) {
         # <TR><TD>33<TD>14<TD>462<TD>2.19017<TD>-106.625
         chomp ();
         my $line = $_;
         if ($line =~ m|<TR><TD>.+<TD>.+<TD>.+<TD>(.+?)<TD>|) {
            my $rate = $1;
            push @triplet, $rate;
            $record .= "\t$rate";
         }
      }
      # Heuristic rule to reject a progres curve:
      # if either the second or third segment has negative slope.
      my $accept = "+";
      if ($triplet[1] < 0 || $triplet[2] < 0) {
         $accept = "-";
      }
      $recordAccept .= "$accept\t";
   }
   $main::recordAccept{$label} = $recordAccept;
}
 
#-------------------------------------------------------
sub execute_script
{   
   my ($script) = @_;
   my $cwd = cwd ();
   $script =~ s|$main::ini{'InputDirectory'}|$main::ini{'ProjectDirectory'}|;
   my $log = $script;
   $log =~ s|\.txt$|.log|;

   chdir ($main::ini{'DynaFitDirectory'});
   my $command = "$main::ini{'DynaFitExe'} $script >>$log";
   my $screen = $script;
   $screen =~ s|$main::ini{'ProjectDirectory'}/||;
   $screen =~ s|/$main::ini{'ScriptFile'}||;
   $screen =~ s|/| : |;
   print $screen, " ... ";
   my $status = system ($command);
   print " status = $status\n";
   chdir ($cwd)
}

#-------------------------------------------------------
sub write_script
{
   my ($path) = @_;
      
   $path =~ s|/[^/]+$||;
   $path =~ s|/data||;

   my $outputDir = $path;
   $outputDir =~ s|$main::ini{'InputDirectory'}|$main::ini{'ProjectDirectory'}|;
   $outputDir .= "/output/$main::ini{'ScriptFile'}";
   $outputDir =~ s|\.txt$||;
   
   my $dataDir = $path;
   $dataDir =~ s|$main::ini{'InputDirectory'}|$main::ini{'ProjectDirectory'}|;
   $dataDir .= "/data";

   $path .= "/$main::ini{'ScriptFile'}";

   my @script = @main::scriptTemplate;
   my $line;
   foreach $line (@script) {

      $line =~ s|OUTPUTDIR|$outputDir|;
      $line =~ s|DATADIR|$dataDir|;
   }

   open (OUT, ">$path") or error_exit ("could not open file\n[$path]!\n");
   print OUT @script;
   close (OUT);   
   return $path;
}

#-------------------------------------------------------
#-------------------------------------------------------
sub initialize
{
	my ($line, $inifile, $logfile);

	if (@ARGV) {
		$inifile = $ARGV[0];
	}
	else {
		$inifile = "$main::program.ini";
	}
	read_inifile ($inifile);

	$logfile = "$main::program.log";
	open (LOG, ">$logfile") or error_exit ("could not open log file\n[$logfile]\n");
}

#-------------------------------------------------------
sub read_inifile
{
	my ($inifile) = @_;
	my ($line, $lineRaw, $key, $value, $contin);

	open (INI, $inifile) or error_exit ("could not open initialization file\n[$inifile]\n");

	$contin = 0;
	$line = "";

	lineRaw:
	foreach $lineRaw (<INI>) {
		chomp ($lineRaw);
		
		# Allow for continuation lines:

		if ($lineRaw =~ m|\\$|) {
			$contin = 1;
		   $lineRaw =~ s|\\$||;
			$line .= $lineRaw;
			next lineRaw;
		}
		else {
			if ($contin) {
				$line .= $lineRaw;
			}
			else {
				$line = $lineRaw;
			}
			$contin = 0;
		}

		$line =~ s|\;.+||g;
		if (!($line =~ m|=|)) {
			next;
		}

		($key, $value) = split /=/, $line;
		$key =~ s|^\s+||;
		$key =~ s|\s+$||;
		$value =~ s|^\s+||;
		$value =~ s|\s+$||;
		$main::ini{$key} = $value;

		$line = "";
	}
	close (INI);
}

#-------------------------------------------------
sub find_files_dir_ext
{
	my ($directory, $extension) = @_;

	undef @main::files;

	$main::match = "";
	if (length($extension) > 0) {
		$main::match = "\.$extension";
	}
	find (\&wanted_files, $directory);

	return @main::files;
}

#-------------------------------------------------
sub find_all_files
{
	my (@tmp);

	@tmp = find_file_extension ("$main::ini{'InputDirectory'}", "", "");

	return @tmp;
}

#-------------------------------------------------
sub find_files
{
	my (@tmp);

	@tmp = find_file_extension ("$main::ini{'InputDirectory'}", "", $main::ini{'FileExtension'});

	return @tmp;
}

#-------------------------------------------------
sub find_file_extension
{
	my ($directory, $file, $extension) = @_;

	undef @main::files;

	$main::match = "";
	if (defined($file) && length($file) > 0) {
		$main::match = $file;
	}
	elsif (length($extension) > 0) {
		$main::match = "\.$extension";
	}

	find (\&wanted_files, $directory);

	return @main::files;
}

#-------------------------------------------------
sub wanted_files
{
	m|$main::match$|i && push @main::files, $File::Find::name;
}

#-------------------------------------------------------
sub terminate
{
	close (LOG);
	if ((defined $main::ini{'PressEnterToFinish'}) && ($main::ini{'PressEnterToFinish'} eq 'n')) {
	   return;
	}
	print "All done.  Press ENTER to terminate.";
	<STDIN>;
}

#-------------------------------------------------------
sub error_exit
{
   my ($message) = @_;
   print "ERROR: $message\n\nPress [Enter] to terminate.";
   <STDIN>;
   exit ();   
}
