#!/usr/bin/perl

use Cwd;
use File::Find;
use File::Copy;
use File::Path qw(make_path);

use strict;
package main;

($main::program = $0) =~ s,.*[\\/],,;
$main::program =~ s|\.pl||;
$main::program =~ s|\.exe||;

$| = 1;  # turn off buffering of STDOUT output

initialize ();

my $inputDir = "$main::ini{'DynaFitDirectory'}/$main::ini{'ProjectDirectory'}";
$inputDir =~ s|\.\.\.|\.\.|;
$main::ini{'InputDirectory'} = $inputDir;

find_files ();
filter_files ();
create_report_files ();
run ();
close_report_files ();
terminate ();

#-------------------------------------------------------
sub run
{
   print <<_END;

============================================================
$main::ini{'Title'}
============================================================

_END

   read_script_template ();
   process_files ();
   write_report ();
   write_parameters_file ();
}

#-------------------------------------------------------
sub write_report
{
   write_report_txt ();
   write_report_csv ();
   write_report_htm ();
}

#-------------------------------------------------------
sub write_report_txt
{
   print TXT <<_END;
MODEL PARAMETERS PER REPLICATE
_END
   foreach (sort keys %main::params) {
      my $key = $_;
      my $record = $main::params{$key};
      $key =~ s|\t|/|;
      print TXT <<_END;

$key
_END
      my @paramNames = split /,/, $main::ini{'Parameters'};
      my @paramUnits = split /,/, $main::ini{'ParameterUnits'};
      print TXT "[I],$main::ini{'ConcentrationUnit'}";
      foreach (@paramNames) {
         my $name = $_;
         my $unit = shift @paramUnits;
         print TXT "\t$name,$unit\t\tCV%"
      }
      print TXT "\n";
      print TXT $record;
   }
}

#-------------------------------------------------------
sub write_report_csv
{
   foreach (sort keys %main::params) {
      my $key = $_;
      my $record = $main::params{$key};
      $key =~ s|\t|/|;
      print CSV <<_END;
$key
_END
      my @paramNames = split /,/, $main::ini{'Parameters'};
      my @paramUnits = split /,/, $main::ini{'ParameterUnits'};
      print CSV "\"[I],$main::ini{'ConcentrationUnit'}\"";
      foreach (@paramNames) {
         my $name = $_;
         my $unit = shift @paramUnits;
         print CSV ",\"$name,$unit\",\"\",\"CV%\""
      }
      print CSV "\n";
      $record =~ s|\t|,|g;
      print CSV $record;
      print CSV "\n";
   }
}

#-------------------------------------------------------
sub write_report_htm
{

   print HTM <<_END;
<h2>$main::ini{'Title'}</h2>

<h3>Model parameters per replicate</h3>
_END

   foreach (sort keys %main::params) {
      my $key = $_;
      my $record = $main::params{$key};
      $key =~ s|\t|/|;
      my $link = "../$main::ini{'DynaFitDirectory'}/$main::ini{'ProjectDirectory'}/$key/output/$main::ini{'ScriptFile'}/index.html";
      $link =~ s|\.txt||;
      print HTM <<_END;
<h4>$key</h4>
[<a href="$link">link</a>]

<p>
<table border="1">
_END
      my @paramNames = split /,/, $main::ini{'Parameters'};
      my @paramUnits = split /,/, $main::ini{'ParameterUnits'};
      my $line = "[I],$main::ini{'ConcentrationUnit'}";
      foreach (@paramNames) {
         my $name = $_;
         my $unit = shift @paramUnits;
         $line .= "\t$name,$unit\t\tCV%";
      }
      $line =~ s|\t|<td class="th">|g;
      print HTM "<tr><td class=\"th\">$line\n";
      
      $record =~ s|\n$||;
      $record = "<tr><td>$record";
      $record =~ s|\n|\n<tr><td>|g;
      $record =~ s|\t|<td>|g;
      print HTM $record;
      print HTM <<_END;
</table>
_END
   }

   my $linkTxt = "$main::ini{'ReportFile'}.txt";
   my $linkCsv = "$main::ini{'ReportFile'}.csv";
   print HTM <<_END;
 <p>
 <hr noshade size="3">
 <p>
 Text export: [<a href="$linkTxt">TXT</a>] [<a href="$linkCsv">CSV</a>]
_END
}

#-------------------------------------------------------
sub write_parameters_file
{
   my $path = $main::ini{'ParametersFile'};
   open (OUT, ">$path") or error_exit ("could not open file\n[$path]\n");
   
   my @paramNames = split /,/, $main::ini{'Parameters'};
   my @paramUnits = split /,/, $main::ini{'ParameterUnits'};
   print OUT "[I]";
   foreach (@paramNames) {
      my $name = $_;
      my $unit = shift @paramUnits;
      print OUT "\t$name,$unit\t\tCV%"
   }
   print OUT "\n\n"; 
   foreach (sort keys %main::params) {
      my $key = $_;      
      my $record = $main::params{$key};
      $key =~ s|/|\t|;
      print OUT "$key\n$record\n";
   }
}

#-------------------------------------------------------
sub create_link
{
   my ($key) = @_;
   my ($compound, $replicate) = split /\//, $key;
   my $prefix = "../";  # we have to back-track from the "results" directory!   This should be made programmatic.
   
   my $compRep = $compound;
   if ($replicate) {
      $compRep .= "/$replicate";
   }
   
   my $link = "<a href=\"$prefix$main::ini{'InputDirectory'}/$compRep/output/$main::ini{'ScriptFile'}/index.html\">";
   $link =~ s|\.txt||;
   $link .= "$key</a>";
   return $link;
}

#-------------------------------------------------------
sub process_files
{
   foreach (@main::files) {
      my $path = $_;
      process_script ($path);
   }
}

#-------------------------------------------------------
sub process_script
{
   my ($path) = @_;
   read_inhibitor_concentrations ($path);
   my $script = write_script ($path);
   if ($main::ini{'ExecuteDynaFit'}) {
      execute_script ($script);
   }
   if ($main::ini{'CompileReport'}) {
      parse_output_params ($script);
   }
}

#-------------------------------------------------------
sub parse_output_params
{
   my ($script) = @_;
   
   my $iTask = 0;
   my $i = 1;
   for (; $i <= $main::ini{'Columns'}; ++$i) {
      if ($main::includeConc[$i-1] eq "+") {
         $iTask++;
         parse_output_params_task ($script, $iTask, $main::concs[$i-1]);
      }
   }
}

#-------------------------------------------------------
sub parse_output_params_task
{
   my ($script, $iTask, $conc) = @_;
   my $task = sprintf ("%3.3d", $iTask);
   my $outputDir = $script;
   $outputDir =~ s|$main::ini{'ScriptFile'}|output|;
   my $outputSubdir = $main::ini{'ScriptFile'};
   $outputSubdir =~ s|\.txt||;

   $script =~ s|$main::ini{'InputDirectory'}/||;
   $script =~ s|/$main::ini{'ScriptFile'}||;

   my $paramFile = "$outputDir/$outputSubdir/txt/parameters-t${task}-LM.txt";
   my @fields = split /\//, $paramFile;
   my ($compound, $replicate, $key);
   if ($main::ini{'Replicates'} && $main::ini{'Replicates'} > 0) {
      $compound = $fields[-6];
      $replicate = $fields[-5];
      $key = "$compound/$replicate";
   }
   else {
      $compound = $fields[-5];
      $key = "$compound";
   }

   open (IN, $paramFile) or error_exit ("could not open file\n[$paramFile]\n");
   my @txt = <IN>;
   close (IN);

   my @paramsToExtract = split /,/, $main::ini{'Parameters'};

   my $record = "$conc";
   foreach (@txt) {
      chomp ();
      my $line = $_;
      $line =~ s| ||g;
      my @fields = split /\t/, $line;
      foreach (@paramsToExtract) {
         my $parName = $_;
         if ($fields[2] && $fields[2] eq $parName) {
            my $par = $fields[3];
            my $err = 0.0;
            my $cv  = 0.0;
            if ($fields[6]) {
               $par = $fields[6];
               $err = $fields[7];
               $cv  = $fields[8];
            }
            $par = sprintf ("%.4g", $par);
            if ($err ne "?") {
               $err = sprintf ("%.3g", $err);
               $cv = sprintf ("%.3g", $cv);
            }
            $record .= "\t$par\t$err\t$cv";
         }
      }
   }
   $record .= "\n";
   $main::params{$key} .= $record;
}

#-------------------------------------------------------
sub write_script
{
   my ($path) = @_;
   $path =~ s|/[^/]+$||;
   $path =~ s|/data||;

   my $outputDir = $path;
   $outputDir =~ s|$main::ini{'InputDirectory'}|$main::ini{'ProjectDirectory'}|;
   $outputDir .= "/output/$main::ini{'ScriptFile'}";
   $outputDir =~ s|\.txt$||;

   my $dataDir = $path;
   $dataDir =~ s|$main::ini{'InputDirectory'}|$main::ini{'ProjectDirectory'}|;
   $dataDir .= "/data";

   my @fields = split /\//, $path;        
   my ($replicate) = pop @fields;
   my ($compound) = pop @fields;

   $path .= "/$main::ini{'ScriptFile'}";

   my @script = @main::scriptTemplate;
   my $line;
   foreach $line (@script) {
      $line =~ s|OUTPUTDIR|$outputDir|;
      $line =~ s|DATADIR|$dataDir|;
      if ($line =~ m|FIRST_COLUMN|) {
         $line = insert_first_column ($compound, $replicate);
      }
      if ($line =~ m|REMAINING_COLUMNS|) {
         $line = insert_remaining_columns ($compound, $replicate);
      }
   }

   open (OUT, ">$path") or error_exit ("could not open file\n[$path]\n");
   print OUT @script;
   close (OUT);

   return $path;
}

#-------------------------------------------------------
sub insert_first_column
{
   my ($compound, $replicate) = @_;   
   my $columnsText = "";
   my $i;
   for ($i = 1; $i <= $main::ini{'Columns'}; ++$i) {
      if ($main::includeConc[$i-1] eq "+") {
         my $iCol = sprintf ("%3d", $i + 1);
         my $conc = $main::concs[$i-1];
         $columnsText .= "   column $iCol | offset auto ? | conc I = $conc\n";
         last;
      }
   }
   return $columnsText;
}

#-------------------------------------------------------
sub insert_remaining_columns
{
   my ($compound, $replicate) = @_;   
   my $columnsText = "";
   my $i;
   my $first = 1;
   for ($i = 1; $i <= $main::ini{'Columns'}; ++$i) {
      if ($main::includeConc[$i-1] eq "+") {
         if ($first) {
            $first = 0;
         }
         else {
            my $iCol = sprintf ("%3d", $i + 1);
            my $conc = $main::concs[$i-1];
            $columnsText .= "[task] | task = fit | data = progress | [data] | column $iCol | offset auto ? | conc I = $conc\n";
         }
      }
   }
   return $columnsText;
}

#-------------------------------------------------------
sub execute_script
{
   my ($script) = @_;
   my $cwd = cwd ();
   $script =~ s|$main::ini{'InputDirectory'}|$main::ini{'ProjectDirectory'}|;
   my $log = $script;
   $log =~ s|\.txt$|.log|;

   chdir ($main::ini{'DynaFitDirectory'});
   my $command = "$main::ini{'DynaFitExe'} $script >>$log";
   my $screen = $script;
   $screen =~ s|$main::ini{'ProjectDirectory'}/||;
   $screen =~ s|/$main::ini{'ScriptFile'}||;
   $screen =~ s|/| : |;
   print $screen, " ... ";
   my $status = system ($command);
   print " status = $status\n";
   chdir ($cwd)
}

#-------------------------------------------------------
sub read_script_template
{
   my $path = $main::ini{'ScriptTemplate'};
   open (IN, $path) or error_exit ("could not open file\n[$path]\n");
   @main::scriptTemplate = <IN>;
   close (IN);
}

#-------------------------------------------------------
sub create_report_files
{
   my $path = $main::ini{'HtmlHeaderFile'};
   open (IN, $path) or error_exit ("could not open file\n[$path]\n");
   @main::htmlHeader = <IN>;
   close (IN);

   $path = $main::ini{'HtmlFooterFile'};
   open (IN, $path) or error_exit ("could not open file\n[$path]\n");
   @main::htmlFooter = <IN>;
   close (IN);

   my $status = mkdir ($main::ini{'ReportDirectory'});
   
   $path = "$main::ini{'ReportDirectory'}/$main::ini{'ReportFile'}.txt";
   open (TXT, ">$path") or error_exit ("could not open output file\n[$path]\n");
   
   $path = "$main::ini{'ReportDirectory'}/$main::ini{'ReportFile'}.csv";
   open (CSV, ">$path") or error_exit ("could not open output file\n[$path]\n");

   $path = "$main::ini{'ReportDirectory'}/$main::ini{'ReportFile'}.html";
   open (HTM, ">$path") or error_exit ("could not open output file\n[$path]\n");
   
   print HTM @main::htmlHeader; 
}

#-------------------------------------------------------
sub close_report_files
{
   print HTM @main::htmlFooter; 

   close (TXT);
   close (HTM);

   my $path = "$main::ini{'ReportDirectory'}/$main::ini{'ReportFile'}.html";
   my $command = "start $path";
   system ($command);
}

#-------------------------------------------------------
sub filter_files
{
   my @files;
   foreach (@main::files) {
      my $path = $_;
      if ($path =~ m|$main::ini{'SheetFileName'}\.$main::ini{'FileExtension'}|) {
         push @files, $path;
      }
   }
   @main::files = @files;
}

#-------------------------------------------------------
sub conc_scale_factor 
{
   my ($unit) = @_;
   my $mult = 1;
   if ($unit =~ m|pM|) {
      $mult = 1.e-12;
   }
   elsif ($unit =~ m|nM|) {
      $mult = 1.e-9;
   }
   elsif ($unit =~ m|uM|) {
      $mult = 1.e-6;
   }
   elsif ($unit =~ m|mM|) {
      $mult = 1.e-3;
   }
   elsif ($unit =~ m|M|) {
      $mult = 1;
   }
   else {
      print "ERROR: wrong concentration unit.\n",
            "Allowable units are M, mM, uM, nM, and pM.";
      termninate ();
   }
   return $mult;
}

#-------------------------------------------------------
sub read_inhibitor_concentrations
{
   my ($path) = @_;

   open (IN, $path) or error_exit ("could not open file\n[$path]\n");
   my @txt = <IN>;
   close (IN);
      
   # We must have a line starting with "[I],UNIT"
   # where UNIT is nM, uM, or mM.
   #
   undef @main::concs;
   undef @main::includeConc;
   foreach (@txt) {
      if (/^\[I\]/) {
         chomp ();
         my @fields = split /\t/;         
         my $header = shift @fields;
         my ($label, $unit) = split ",", $header;
         my $multInput = conc_scale_factor ($unit);
         my $multConc = conc_scale_factor ($main::ini{'ConcentrationUnit'});
         my $mult = $multInput/$multConc;        
         my $conc;
         foreach $conc (@fields) {
            $conc *= $mult;
         }
         @main::concs = @fields;
      }
      elsif (/INCLUDE/) {
         chomp ();
         my @fields = split /\t/;
         shift @fields;
         @main::includeConc = @fields;
      }
   }
   # If we didn't find the INCLUDE line, assume that all concentrations should
   # be included in the analysis.
   #
   if (!@main::includeConc) {
      my $n = scalar (@main::concs);
      my $i = 0;
      for (; $i < $n; ++$i) {
         push @main::includeConc, '+';
      }
   }
}

#-------------------------------------------------------
#-------------------------------------------------------
sub initialize
{
	my ($line, $inifile, $logfile);

	if (@ARGV) {
		$inifile = $ARGV[0];
	}
	else {
		$inifile = "$main::program.ini";
	}
	read_inifile ($inifile);

	$logfile = "$main::program.log";
	open (LOG, ">$logfile") or error_exit ("could not open log file\n[$logfile]\n");
}

#-------------------------------------------------------
sub read_inifile
{
	my ($inifile) = @_;
	my ($line, $lineRaw, $key, $value, $contin);

	open (INI, $inifile) or error_exit ("could not open initialization file\n[$inifile]\n");

	$contin = 0;
	$line = "";

	lineRaw:
	foreach $lineRaw (<INI>) {
		chomp ($lineRaw);
		
		# Allow for continuation lines:

		if ($lineRaw =~ m|\\$|) {
			$contin = 1;
		   $lineRaw =~ s|\\$||;
			$line .= $lineRaw;
			next lineRaw;
		}
		else {
			if ($contin) {
				$line .= $lineRaw;
			}
			else {
				$line = $lineRaw;
			}
			$contin = 0;
		}

		$line =~ s|\;.+||g;
		if (!($line =~ m|=|)) {
			next;
		}

		($key, $value) = split /=/, $line;
		$key =~ s|^\s+||;
		$key =~ s|\s+$||;
		$value =~ s|^\s+||;
		$value =~ s|\s+$||;
		$main::ini{$key} = $value;

		$line = "";
	}
	close (INI);
}

#-------------------------------------------------
sub find_files_dir_ext
{
	my ($directory, $extension) = @_;

	undef @main::files;

	$main::match = "";
	if (length($extension) > 0) {
		$main::match = "\.$extension";
	}
	find (\&wanted_files, $directory);

	return @main::files;
}

#-------------------------------------------------
sub find_all_files
{
	my (@tmp);

	@tmp = find_file_extension ("$main::ini{'InputDirectory'}", "", "");

	return @tmp;
}

#-------------------------------------------------
sub find_files
{
	my (@tmp);

	@tmp = find_file_extension ("$main::ini{'InputDirectory'}", "", $main::ini{'FileExtension'});

	return @tmp;
}

#-------------------------------------------------
sub find_file_extension
{
	my ($directory, $file, $extension) = @_;

	undef @main::files;

	$main::match = "";
	if (defined($file) && length($file) > 0) {
		$main::match = $file;
	}
	elsif (length($extension) > 0) {
		$main::match = "\.$extension";
	}

	find (\&wanted_files, $directory);

	return @main::files;
}

#-------------------------------------------------
sub wanted_files
{
	m|$main::match$|i && push @main::files, $File::Find::name;
}

#-------------------------------------------------------
sub terminate
{
	close (LOG);
	if ((defined $main::ini{'PressEnterToFinish'}) && ($main::ini{'PressEnterToFinish'} eq 'n')) {
	   return;
	}
	print "All done.  Press ENTER to terminate.";
	<STDIN>;
}

#-------------------------------------------------------
sub error_exit
{
   my ($message) = @_;
   print "ERROR: $message\n\nPress [Enter] to terminate.";
   <STDIN>;
   exit ();   
}

