#!/usr/bin/perl

use Cwd;
use File::Find;
use File::Copy;
use File::Path qw(make_path);

use strict;
package main;

($main::program = $0) =~ s,.*[\\/],,;
$main::program =~ s|\.pl||;
$main::program =~ s|\.exe||;

$| = 1;  # turn off buffering of STDOUT output

initialize ();

my $inputDir = "$main::ini{'DynaFitDirectory'}/$main::ini{'ProjectDirectory'}";
$inputDir =~ s|\.\.\.|\.\.|;
$main::ini{'InputDirectory'} = $inputDir;

find_files ();
filter_files ();
create_report_files ();
run ();
close_report_files ();
terminate ();

#-------------------------------------------------------
sub run
{
   print <<_END;

============================================================
$main::ini{'Title'}
============================================================

_END

   read_script_template ();
   process_files ();
   write_report ();
   write_parameters_file ();
}

#-------------------------------------------------------
sub write_report
{
   write_report_txt ();
   write_report_csv ();
   write_report_htm ();
}

#-------------------------------------------------------
sub write_report_txt
{
   my @paramNames = split /,/, $main::ini{'Parameters'};
   my @paramUnits = split /,/, $main::ini{'ParameterUnits'};
   
   print TXT "Compound";
   foreach (@paramNames) {
      my $name = $_;
      my $unit = shift @paramUnits;
      print TXT "\t$name, $unit\t\tcv%"
   }
   print TXT "\n";

   for (sort keys %main::params) {
      my $key = $_;
      my $record = $main::params{$key};
      if (!$record) {
         next;
      }
      print TXT "$key\t$record\n";
   }
}

#-------------------------------------------------------
sub write_report_csv
{
   my @paramNames = split /,/, $main::ini{'Parameters'};
   my @paramUnits = split /,/, $main::ini{'ParameterUnits'};

   print CSV "Compound";
   foreach (@paramNames) {
      my $name = $_;
      my $unit = shift @paramUnits;
      print CSV ",\"$name, $unit\",,cv%"
   }
   print CSV "\n";
   
   for (sort keys %main::params) {
      my $key = $_;
      my $record = $main::params{$key};
      if (!$record) {
         next;
      }
      $record =~ s|\t|,|g;
      print CSV "$key,$record\n";
   }
}

#-------------------------------------------------------
sub write_report_htm
{
   my @paramNames = split /,/, $main::ini{'Parameters'};
   my @paramUnits = split /,/, $main::ini{'ParameterUnits'};

   print HTM <<_END;
<h2>$main::ini{'Title'}</h2>
<table border="1">
<tr>
<td class="th">Compound
_END

   foreach (@paramNames) {
      my $name = $_;
      my $unit = shift @paramUnits;
      print HTM<<_END
<td class="th">$name,$unit
<td class="th">
<td class="th">CV%
_END
   }
   print HTM "\n";

   for (sort keys %main::params) {
      my $key = $_;
      my $link = create_link ($key);
      my $record = $main::params{$key};
      if (!$record) {
         next;
      }
      $record =~ s|\t|\t<td>|g;
      $record =~ s|<td>$||;
      $record =~ s|td|td align=\"right\"|g;
      print HTM "<tr><td>$link<td align=\"right\">$record\n";
   }
   print HTM <<_END;
</table>
_END
   my $linkTxt = "$main::ini{'ReportFile'}.txt";
   my $linkCsv = "$main::ini{'ReportFile'}.csv";
   print HTM <<_END;
 <p>
 <hr noshade size="3">
 <p>
 Text export: [<a href="$linkTxt">TXT</a>] [<a href="$linkCsv">CSV</a>]
_END
}

#-------------------------------------------------------
sub write_parameters_file
{
   my $path = $main::ini{'ParametersFile'};
   open (OUT, ">$path") or error_exit ("could not open output file\n[$path]\n");
   
   my @paramNames = split /,/, $main::ini{'Parameters'};
   my @paramUnits = split /,/, $main::ini{'ParameterUnits'};
   
   print OUT "Compound";
   foreach (@paramNames) {
      my $name = $_;
      my $unit = shift @paramUnits;
      print OUT "\t$name, $unit\t\tcv%"
   }
   print OUT "\n";

   for (sort keys %main::params) {
      my $key = $_;
      my $record = $main::params{$key};
      if (defined ($record)) {
         print OUT "$key\t$record\n";
      }
   }
   close (OUT);
}

#-------------------------------------------------------
sub create_link
{
   my ($key) = @_;
   my ($compound) = $key;
   my $prefix = "../";  # we have to back-track from the "results" directory!   This should be made programmatic.
   
   my $link = "<a href=\"$prefix$main::ini{'InputDirectory'}/$compound/output/$main::ini{'ScriptFile'}/index.html\">";
   if (!$main::ini{'UseReplicates'}) {
      $link = "<a href=\"$prefix$main::ini{'InputDirectory'}/$compound/output/$main::ini{'ScriptFile'}/index.html\">";
   }
   $link =~ s|\.txt||;
   $link .= "$key</a>";
   return $link;
}

#-------------------------------------------------------
sub process_files
{
   foreach (@main::files) {
      my $path = $_;
      process_script ($path);
   }
}

#-------------------------------------------------------
sub process_script
{
   my ($path) = @_;
   my $script = "";
   if ($main::ini{'UseReplicates'}) {
      my $subdir = sprintf ("%s%d", "R", $main::ini{'Replicates'});
      if ($path =~ m|/$subdir/|) {
         $path =~ s|/$subdir||;
         $script = write_script ($path);
      }
      else {
         return;
      }
   }
   else {
      $script = write_script ($path);
   }
   if ($main::ini{'ExecuteDynaFit'}) {
      execute_script ($script);
   }
   if ($main::ini{'CompileReport'}) {
      parse_output_params ($script, 1);
      if ($main::ini{'ModelDiscrimination'}) {
         parse_output_params ($script, 2);
      }
   }
}

#-------------------------------------------------------
sub extract_params
{
   my ($path) = @_;
   open (IN, $path) or error_exit ("could not open file\n[$path]\n");
   my @txt = <IN>;
   close (IN);
   my @paramsToExtract = split /,/, $main::ini{'Parameters'};
   my @paramValues;
   foreach (@txt) {
      chomp ();
      my $line = $_;
      $line =~ s| ||g;
      my @fields = split /\t/, $line;
      foreach (@paramsToExtract) {
         my $par = $_;
         if ($fields[2] && $fields[2] eq $par) {
            push @paramValues, $fields[6];
         }
      }
   }
   return @paramValues;
}

#-------------------------------------------------------
sub parse_output_params
{
   my ($script, $iTask) = @_;
   my $task = sprintf ("%3.3d", $iTask);
   my $outputDir = $script;
   $outputDir =~ s|$main::ini{'ScriptFile'}|output|;
   my $outputSubdir = $main::ini{'ScriptFile'};
   $outputSubdir =~ s|\.txt||;

   $script =~ s|$main::ini{'InputDirectory'}/||;
   $script =~ s|/$main::ini{'ScriptFile'}||;

   my $paramFile = "$outputDir/$outputSubdir/txt/parameters-t${task}-LM.txt";
   my @fields = split /\//, $paramFile;
   my $compound = $fields[-5];
   if (!$main::ini{'UseReplicates'}) {
      # $compound = $fields[-6];
   }

   open (IN, $paramFile) or error_exit ("could not open file\n[$paramFile]\n");
   my @txt = <IN>;
   close (IN);

   my @paramsToExtract = split /,/, $main::ini{'Parameters'};

   my $record = "";
   foreach (@txt) {
      chomp ();
      my $line = $_;
      $line =~ s| ||g;
      my @fields = split /\t/, $line;
      foreach (@paramsToExtract) {
         my $parName = $_;
         if ($fields[2] && ($fields[2] eq $parName) && (($fields[1] eq "*") || ($fields[1] eq "1"))) {
            my $par = $fields[3];
            my $err = 0.0;
            my $cv  = 0.0;
            if ($fields[6]) {
               $par = $fields[6];
               $err = $fields[7];
               $cv  = $fields[8];
            }
            $par = sprintf ("%.4g", $par);
            if ($err ne "?") {
               $err = sprintf ("%.3g", $err);
               $cv = sprintf ("%.3g", $cv);
            }
            $record .= "$par\t$err\t$cv\t";
         }
      }
   }     
   if ($iTask == 1) {
      my $key = "$compound/1";
      $main::params{$key} = $record;
   }
   else {
      my $record1 = $main::params{"$compound/1"};
      my $record2 = $record;
            
      my @fields1 = split /\t/, $record1;
      my @fields2 = split /\t/, $record2;
            
      my $en1   = $fields1[3];
      my $en1se = $fields1[4];
      my $cvki1 = $fields1[2];
      
      if ($cvki1 > 100 || 
          $en1 > $main::ini{'EnzymeMax'} || 
          $en1 < $main::ini{'EnzymeMin'} || 
          $en1se eq "?") {         
         $record = $record2;
      }
      else {
         $record = $record1;
      }
      undef $main::params{"$compound/1"};
      my $key = "$compound";
      $main::params{$key} = $record;
   }
}

#-------------------------------------------------------
sub compile_rate_files
{
   my ($path) = @_;
   my $rateFiles = "";
   if (!$main::ini{'UseReplicates'}) {
      $rateFiles = "   file data/$main::ini{'RateFile'}\n";      
   }
   else {
      my $iRep = 1;
      for (; $iRep <= $main::ini{'Replicates'}; ++$iRep) {
         my $rep = sprintf ("%s%d", "R", $iRep);
         $rateFiles .= "   file $rep/data/$main::ini{'RateFile'} | label repl $iRep\n";      
      }
   }
   return $rateFiles;
}

#-------------------------------------------------------
sub write_script
{
   my ($path) = @_;
   $path =~ s|/[^/]+$||;
   $path =~ s|/data||;

   my $outputDir = $path;
   $outputDir =~ s|$main::ini{'InputDirectory'}|$main::ini{'ProjectDirectory'}|;
   $outputDir .= "/output/$main::ini{'ScriptFile'}";
   $outputDir =~ s|\.txt$||;

   my $dataDir = $path;
   $dataDir =~ s|$main::ini{'InputDirectory'}|$main::ini{'ProjectDirectory'}|;

   my $rateFiles = compile_rate_files ($path);
   $path .= "/$main::ini{'ScriptFile'}";
   
   my @script = @main::scriptTemplate;
   my $line;
   foreach $line (@script) {
      $line =~ s|OUTPUTDIR|$outputDir|;
      $line =~ s|DATADIR|$dataDir|;
      $line =~ s|RATEFILES|$rateFiles|;
   }

   open (OUT, ">$path") or error_exit ("could not open file\n[$path]\n");
   print OUT @script;
   close (OUT);

   return $path;
}

#-------------------------------------------------------
sub execute_script
{
   my ($script) = @_;
   my $cwd = cwd ();
   $script =~ s|$main::ini{'InputDirectory'}|$main::ini{'ProjectDirectory'}|;
   my $log = $script;
   $log =~ s|\.txt$|.log|;

   chdir ($main::ini{'DynaFitDirectory'});
   my $command = "$main::ini{'DynaFitExe'} $script >>$log";
   my $screen = $script;
   $screen =~ s|$main::ini{'ProjectDirectory'}/||;
   $screen =~ s|/$main::ini{'ScriptFile'}||;
   $screen =~ s|/| : |;
   print $screen, " ... ";
   my $status = system ($command);
   print " status = $status\n";
   chdir ($cwd)
}

#-------------------------------------------------------
sub read_script_template
{
   my $path = $main::ini{'ScriptTemplate'};
   open (IN, $path) or error_exit ("could not open file\n[$path]\n");
   @main::scriptTemplate = <IN>;
   close (IN);
}

#-------------------------------------------------------
sub create_report_files
{
   my $path = $main::ini{'HtmlHeaderFile'};
   open (IN, $path) or error_exit ("could not open file\n[$path]\n");
   @main::htmlHeader = <IN>;
   close (IN);

   $path = $main::ini{'HtmlFooterFile'};
   open (IN, $path) or error_exit ("could not open file\n[$path]\n");
   @main::htmlFooter = <IN>;
   close (IN);

   my $status = mkdir ($main::ini{'ReportDirectory'});
   
   $path = "$main::ini{'ReportDirectory'}/$main::ini{'ReportFile'}.txt";
   open (TXT, ">$path") or error_exit ("could not open output file\n[$path]\n");
   
   $path = "$main::ini{'ReportDirectory'}/$main::ini{'ReportFile'}.csv";
   open (CSV, ">$path") or error_exit ("could not open output file\n[$path]\n");
   
   $path = "$main::ini{'ReportDirectory'}/$main::ini{'ReportFile'}.html";
   open (HTM, ">$path") or error_exit ("could not open output file\n[$path]\n");
   
   print HTM @main::htmlHeader; 
}

#-------------------------------------------------------
sub close_report_files
{
   print HTM @main::htmlFooter; 

   close (TXT);
   close (CSV);
   close (HTM);

   my $path = "$main::ini{'ReportDirectory'}/$main::ini{'ReportFile'}.html";
   my $command = "start $path";
   system ($command);
}

#-------------------------------------------------------
sub filter_files
{
   my @files;
   foreach (@main::files) {
      my $path = $_;
      if ($path =~ m|$main::ini{'SheetFileName'}\.$main::ini{'FileExtension'}|) {
         push @files, $path;
      }
   }
   @main::files = @files;
}

#-------------------------------------------------------
#-------------------------------------------------------
sub initialize
{
	my ($line, $inifile, $logfile);

	if (@ARGV) {
		$inifile = $ARGV[0];
	}
	else {
		$inifile = "$main::program.ini";
	}
	read_inifile ($inifile);

	$logfile = "$main::program.log";
	open (LOG, ">$logfile") or error_exit ("could not open log file\n[$logfile]\n");
}

#-------------------------------------------------------
sub read_inifile
{
	my ($inifile) = @_;
	my ($line, $lineRaw, $key, $value, $contin);

	open (INI, $inifile) or error_exit ("could not open initialization file\n[$inifile]\n");

	$contin = 0;
	$line = "";

	lineRaw:
	foreach $lineRaw (<INI>) {
		chomp ($lineRaw);
		
		# Allow for continuation lines:

		if ($lineRaw =~ m|\\$|) {
			$contin = 1;
		   $lineRaw =~ s|\\$||;
			$line .= $lineRaw;
			next lineRaw;
		}
		else {
			if ($contin) {
				$line .= $lineRaw;
			}
			else {
				$line = $lineRaw;
			}
			$contin = 0;
		}

		$line =~ s|\;.+||g;
		if (!($line =~ m|=|)) {
			next;
		}

		($key, $value) = split /=/, $line;
		$key =~ s|^\s+||;
		$key =~ s|\s+$||;
		$value =~ s|^\s+||;
		$value =~ s|\s+$||;
		$main::ini{$key} = $value;

		$line = "";
	}
	close (INI);
}

#-------------------------------------------------
sub find_files_dir_ext
{
	my ($directory, $extension) = @_;

	undef @main::files;

	$main::match = "";
	if (length($extension) > 0) {
		$main::match = "\.$extension";
	}
	find (\&wanted_files, $directory);

	return @main::files;
}

#-------------------------------------------------
sub find_all_files
{
	my (@tmp);

	@tmp = find_file_extension ("$main::ini{'InputDirectory'}", "", "");

	return @tmp;
}

#-------------------------------------------------
sub find_files
{
	my (@tmp);

	@tmp = find_file_extension ("$main::ini{'InputDirectory'}", "", $main::ini{'FileExtension'});

	return @tmp;
}

#-------------------------------------------------
sub find_file_extension
{
	my ($directory, $file, $extension) = @_;

	undef @main::files;

	$main::match = "";
	if (defined($file) && length($file) > 0) {
		$main::match = $file;
	}
	elsif (length($extension) > 0) {
		$main::match = "\.$extension";
	}

	find (\&wanted_files, $directory);

	return @main::files;
}

#-------------------------------------------------
sub wanted_files
{
	m|$main::match$|i && push @main::files, $File::Find::name;
}

#-------------------------------------------------------
sub terminate
{
	close (LOG);
	if ((defined $main::ini{'PressEnterToFinish'}) && ($main::ini{'PressEnterToFinish'} eq 'n')) {
	   return;
	}
	print "All done.  Press ENTER to terminate.";
	<STDIN>;
}

#-------------------------------------------------------
sub error_exit
{
   my ($message) = @_;
   print "ERROR: $message\n\nPress [Enter] to terminate.";
   <STDIN>;
   exit ();   
}

