#!/usr/bin/perl

use Cwd;
use File::Find;
use File::Copy;
use File::Path qw(make_path);

use strict;
package main;

($main::program = $0) =~ s,.*[\\/],,;
$main::program =~ s|\.pl||;
$main::program =~ s|\.exe||;

$| = 1;  # turn off buffering of STDOUT output

initialize ();

my $inputDir = "$main::ini{'DynaFitDirectory'}/$main::ini{'ProjectDirectory'}";
$inputDir =~ s|\.\.\.|\.\.|;
$main::ini{'InputDirectory'} = $inputDir;

find_files ();
filter_files ();
create_report_files ();
run ();
close_report_files ();
terminate ();

#-------------------------------------------------------
sub run
{
   print <<_END;

=======================================================
$main::ini{'Title'}
=======================================================

_END

   read_script_template ();
   read_estimated_kis ();
   read_estimated_offsets ();
   process_files ();
   if (!$main::ini{'ExecuteDynaFit'} && !$main::ini{'CompileReport'}) {
      return;
   }
   compute_averages ();
   write_report ();
   write_parameters_file ();
}

#-------------------------------------------------------
sub read_estimated_offsets
{
   my $path = $main::ini{'OffsetEstimateFile'};
   open (IN, $path) or error_exit ("could not open [$path]!\n");
   my @txt = <IN>;
   close (IN);
   
   shift @txt;   # throw away the first (header line)
   
   my $key = "";
   my $record = "";
   line:
   foreach (@txt) {
      my $line = $_;
      chomp ($line);
      my @fields = split /\t/, $line;
      my $n = scalar (@fields);
      if ($n < 2) {
         next line;
      }
      elsif ($n == 2) {
         if ($record ne "" && $key ne "") {
            $main::algebOffsets{$key} = $record;
         }
         $key = "$fields[0]/$fields[1]";
         $record = "";
      }
      else {
         my $offset = $fields[$main::ini{'OffsetEstimateField'}];
         $record .= "$offset\t";
      }
   }
   $main::algebOffsets{$key} = $record;
}

#-------------------------------------------------------
sub read_estimated_kis
{
   my $path = $main::ini{'KiAppEstimateFile'};
   open (IN, $path) or error_exit ("could not open [$path]!\n");
   my @txt = <IN>;
   close (IN);

   shift @txt;   # throw away the first (header line)
   
   foreach (@txt) {
      my $line = $_;
      chomp ($line);
      my @fields = split /\t/, $line;
      my $compound = $fields[0];
      my $Ki = $fields[$main::ini{'KiAppEstimateField'}];
      $main::estimatedKi{$compound} = $Ki;
   }
}

#-------------------------------------------------------
sub compute_averages
{
   if ($main::ini{'AveragesGeometric'}) {
      compute_averages_geometric ();
   }
   else {
      compute_averages_arithmetic ();
   }
}

#-------------------------------------------------------
sub compute_averages_geometric
{
   #
   # Compute GEOMETRIC means of rate constants, i.e.,
   # ARITHMETIC means of LOGARITHMS of rate constants.
   #
   my ($ksubAv, $kdIAv, $kinactAv, $KiAv, $kinactKiAv);
   my $ln10 = log(10.0);
   my $nAveraged = 0;

   foreach (sort keys %main::resultsReplicates) {
      my $key = $_;
      my ($compound, $replicate) = split /\t/, $key;
      my $record = $main::resultsReplicates{$key};
      my @fields = split /\t/, $record;
      
      my $n = @fields;
      
      my ($ksub, $kdI, $kinact, $Ki, $kinactKi);

      if ($replicate eq 'R1') {
         $ksubAv = $kdIAv = $kinactAv = $KiAv = $kinactKiAv = 0;
      }
      #
      # Average only values that are associated with less than 100% CV.
      #
      if ($fields[2+2] < 100 && $fields[5+2] < 100 && $fields[8+2] < 100) {
         $nAveraged++;
         $ksubAv        += log($fields[2])/$ln10;
         $kdIAv         += log($fields[5])/$ln10;
         $kinactAv      += log($fields[8])/$ln10;
         $KiAv          += log($fields[11])/$ln10;
         $kinactKiAv    += log($fields[14])/$ln10;
      }
            
      if ($replicate eq 'R3') {
         $ksubAv        = sprintf ("%g", $ksubAv / $nAveraged);
         $kdIAv         = sprintf ("%g", $kdIAv / $nAveraged);
         $kinactAv      = sprintf ("%g", $kinactAv / $nAveraged);
         $KiAv          = sprintf ("%g", $KiAv / $nAveraged);
         $kinactKiAv    = sprintf ("%g", $kinactKiAv / $nAveraged);
         my $recordAv = join "\t", ($ksubAv, $kdIAv, $kinactAv, $KiAv, $kinactKiAv);
         $main::resultsAver{$compound} = $recordAv;
         $nAveraged = 0;
      }
   }

   my ($ksubSd, $kdISd, $kinactSd, $KiSd, $kinactKiSd);
   $nAveraged = 0;

   foreach (sort keys %main::resultsReplicates) {
      my $key = $_;
      my ($compound, $replicate) = split /\t/, $key;
      my $record = $main::resultsReplicates{$key};
      my @fields = split /\t/, $record;
      
      my $n = @fields;

      my ($ksub, $kdI, $kinact, $Ki, $kinactKi);
      
      if ($replicate eq 'R1') {
         $ksubSd = $kdISd = $kinactSd = $KiSd = $kinactKiSd = 0;
      }
      #
      # Average only values that are associated with less than 100% CV.
      #
      if ($fields[2+2] < 100 && $fields[5+2] < 100 && $fields[8+2] < 100) {
         $nAveraged++;
         $ksub          = log($fields[2])/$ln10;
         $kdI           = log($fields[5])/$ln10;
         $kinact        = log($fields[8])/$ln10;
         $Ki            = log($fields[11])/$ln10;
         $kinactKi      = log($fields[14])/$ln10;

         my ($ksubAv, $kdIAv, $kinactAv, $KiAv, $kinactKiAv) = split "\t", $main::resultsAver{$compound};

         $ksubSd        += ($ksub - $ksubAv)*($ksub - $ksubAv);
         $kdISd         += ($kdI - $kdIAv)*($kdI - $kdIAv);
         $kinactSd      += ($kinact - $kinactAv)*($kinact - $kinactAv);
         $KiSd          += ($Ki - $KiAv)*($Ki - $KiAv);
         $kinactKiSd    += ($kinactKi - $kinactKiAv)*($kinactKi - $kinactKiAv);
      }
      
      if ($replicate eq 'R3') {
         if (($nAveraged - 1) > 0) {
            $ksubSd     = sprintf ("%g", sqrt($ksubSd / ($nAveraged - 1)));
            $kdISd      = sprintf ("%g", sqrt($kdISd / ($nAveraged - 1)));
            $kinactSd   = sprintf ("%g", sqrt($kinactSd / ($nAveraged - 1)));
            $KiSd       = sprintf ("%g", sqrt($KiSd / ($nAveraged - 1)));
            $kinactKiSd = sprintf ("%g", sqrt($kinactKiSd / ($nAveraged - 1)));
         }
         else {
            $ksubSd = $kdISd = $kinactSd = $KiSd = $kinactSd = 0;
         }
         my $recordSd = join "\t", ($ksubSd, $kdISd, $kinactSd, $KiSd, $kinactKiSd);
         $main::resultsSdev{$compound} = $recordSd;
         $nAveraged = 0;
      }
   }
}

#-------------------------------------------------------
sub compute_averages_arithmetic
{
   #
   # Compute ARITHMETIC means of rate constants.
   #
   my ($ksubAv, $kdIAv, $kinactAv, $KiAv, $kinactKiAv);
   my $ln10 = log(10.0);
   my $nAveraged = 0;

   foreach (sort keys %main::resultsReplicates) {
      my $key = $_;
      my ($compound, $replicate) = split /\t/, $key;
      my $record = $main::resultsReplicates{$key};
      my @fields = split /\t/, $record;
      
      my $n = @fields;
      
      my ($ksub, $kdI, $kinact, $Ki, $kinactKi);

      if ($replicate eq 'R1') {
         $ksubAv = $kdIAv = $kinactAv = $KiAv = $kinactKiAv = 0;
      }
      #
      # Average only values that are associated with less than 100% CV.
      #
      if ($fields[2+2] < 100 && $fields[5+2] < 100 && $fields[8+2] < 100) {
         $nAveraged++;
         $ksubAv        += $fields[2];
         $kdIAv         += $fields[5];
         $kinactAv      += $fields[8];
         $KiAv          += $fields[11];
         $kinactKiAv    += $fields[14];
      }
            
      if ($replicate eq 'R3') {
         $ksubAv        = sprintf ("%g", $ksubAv / $nAveraged);
         $kdIAv         = sprintf ("%g", $kdIAv / $nAveraged);
         $kinactAv      = sprintf ("%g", $kinactAv / $nAveraged);
         $KiAv          = sprintf ("%g", $KiAv / $nAveraged);
         $kinactKiAv    = sprintf ("%g", $kinactKiAv / $nAveraged);
         my $recordAv = join "\t", ($ksubAv, $kdIAv, $kinactAv, $KiAv, $kinactKiAv);
         $main::resultsAver{$compound} = $recordAv;
         $nAveraged = 0;
      }
   }

   my ($ksubSd, $kdISd, $kinactSd, $KiSd, $kinactKiSd);
   $nAveraged = 0;

   foreach (sort keys %main::resultsReplicates) {
      my $key = $_;
      my ($compound, $replicate) = split /\t/, $key;
      my $record = $main::resultsReplicates{$key};
      my @fields = split /\t/, $record;
      
      my $n = @fields;

      my ($ksub, $kdI, $kinact, $Ki, $kinactKi);
      
      if ($replicate eq 'R1') {
         $ksubSd = $kdISd = $kinactSd = $KiSd = $kinactKiSd = 0;
      }
      #
      # Average only values that are associated with less than 100% CV.
      #
      if ($fields[2+2] < 100 && $fields[5+2] < 100 && $fields[8+2] < 100) {
         $nAveraged++;
         $ksub          = $fields[2];
         $kdI           = $fields[5];
         $kinact        = $fields[8];
         $Ki            = $fields[11];
         $kinactKi      = $fields[14];

         my ($ksubAv, $kdIAv, $kinactAv, $KiAv, $kinactKiAv) = split "\t", $main::resultsAver{$compound};

         $ksubSd        += ($ksub - $ksubAv)*($ksub - $ksubAv);
         $kdISd         += ($kdI - $kdIAv)*($kdI - $kdIAv);
         $kinactSd      += ($kinact - $kinactAv)*($kinact - $kinactAv);
         $KiSd          += ($Ki - $KiAv)*($Ki - $KiAv);
         $kinactKiSd    += ($kinactKi - $kinactKiAv)*($kinactKi - $kinactKiAv);
      }
      
      if ($replicate eq 'R3') {
         if (($nAveraged - 1) > 0) {
            $ksubSd     = sprintf ("%g", sqrt($ksubSd / ($nAveraged - 1)));
            $kdISd      = sprintf ("%g", sqrt($kdISd / ($nAveraged - 1)));
            $kinactSd   = sprintf ("%g", sqrt($kinactSd / ($nAveraged - 1)));
            $KiSd       = sprintf ("%g", sqrt($KiSd / ($nAveraged - 1)));
            $kinactKiSd = sprintf ("%g", sqrt($kinactKiSd / ($nAveraged - 1)));
         }
         else {
            $ksubSd = $kdISd = $kinactSd = $KiSd = $kinactSd = 0;
         }
         my $recordSd = join "\t", ($ksubSd, $kdISd, $kinactSd, $KiSd, $kinactKiSd);
         $main::resultsSdev{$compound} = $recordSd;
         $nAveraged = 0;
      }
   }
}

#-------------------------------------------------------
sub write_report
{
   write_report_txt ();
   write_report_csv ();
   write_report_htm ();
}

#-------------------------------------------------------
sub write_report_txt
{
   my $record = "Cpd\tRepl\tr.m.s.\t";
   $record .= "$main::ini{'kSubstrateLabel'}\t +/- \tcv%\t";
   $record .= "kdI\t +/- \tcv%\t";
   $record .= "kinact\t +/- \tcv%\t";
   $record .= "Ki\t +/- \tcv%\t";
   $record .= "kinact/Ki\t +/- \tcv%\t";
   print TXT $record, "\n";
   
   foreach (sort keys %main::resultsReplicates) {
      my $key = $_;
      my $record = $main::resultsReplicates{$key};
      my $rms = $main::rmsReplicates{$key};
      my @fields = split /\t/, $record;
      print TXT "$fields[0]\t$fields[1]\t$rms\t";
      shift @fields;
      shift @fields;
      print TXT join "\t", @fields;
      print TXT "\n";
   }

   print TXT "\nAVERAGES AND STANDARD DEVIATIONS\n\n";

   $record = "Cpd\t";
   if ($main::ini{'AveragesGeometric'}) {
      $record .= "log($main::ini{'kSubstrateLabel'})\t +/- \tcv%\t";
      $record .= "log(kdI)\t +/- \tcv%\t";
      $record .= "log(kinact)\t +/- \tcv%\t";
      $record .= "log(Ki)\t +/- \tcv%\t";
      $record .= "log(kinact/Ki)\t +/- \tcv%\t";
   }
   else {
      $record .= "$main::ini{'kSubstrateLabel'}\t +/- \tcv%\t";
      $record .= "kdI\t +/- \tcv%\t";
      $record .= "kinact\t +/- \tcv%\t";
      $record .= "Ki\t +/- \tcv%\t";
      $record .= "kinact/Ki\t +/- \tcv%\t";
   }
   print TXT $record, "\n";
   
   foreach (sort keys %main::resultsAver) {
      my $key = $_;
      my @av = split /\t/, $main::resultsAver{$key};
      my @sd = split /\t/, $main::resultsSdev{$key};
      print TXT "$key\t";
      my $n = scalar (@av);
      my $i = 0;
      for (; $i < $n; ++$i) {
         my $av = $av[$i];
         my $sd = $sd[$i];
         my $cv = 0;
         if ($av[$i] != 0) {
            $cv = sprintf ("%.1f", 100*abs($sd[$i]/$av[$i]));
         }
         print TXT "$av\t$sd\t$cv\t";
      }
      print TXT "\n";
   }
}

#-------------------------------------------------------
sub write_report_csv
{
   my $record = "\"Cpd\",\"Repl\",\"r.m.s.\",";
   $record .= "\"log$main::ini{'kSubstrateLabel'}\",\" +/- \",\"cv%\",";
   $record .= "\"kdI\",\" +/- \",\"cv%\",";
   $record .= "\"kinact\",\" +/- \",\"cv%\",";
   $record .= "\"Ki\",\" +/- \",\"cv%\",";
   $record .= "\"kinact/Ki\",\" +/- \",\"cv%\"";
   print CSV $record, "\n";
   
   foreach (sort keys %main::resultsReplicates) {
      my $key = $_;
      my $record = $main::resultsReplicates{$key};
      my $rms = $main::rmsReplicates{$key};
      my @fields = split /\t/, $record;
      print CSV "$fields[0],$fields[1],$rms,";
      shift @fields;
      shift @fields;
      print CSV join ",", @fields;
      print CSV "\n";
   }

   print CSV "\nAVERAGES AND STANDARD DEVIATIONS\n\n";

   $record = "\"Cpd\",";
   if ($main::ini{'AveragesGeometric'}) {
      $record .= "\"log($main::ini{'kSubstrateLabel'})\",\" +/- \",\"cv%\",";
      $record .= "\"log(kdI)\",\" +/- \",\"cv%\",";
      $record .= "\"log(kinact)\",\" +/- \",\"cv%\",";
      $record .= "\"log(Ki)\",\" +/- \",\"cv%\",";
      $record .= "\"log(kinact/Ki)\",\" +/- \",\"cv%\",";
   }
   else {
      $record .= "\"$main::ini{'kSubstrateLabel'}\",\" +/- \",\"cv%\",";
      $record .= "\"kdI\",\" +/- \",\"cv%\",";
      $record .= "\"kinact\",\" +/- \",\"cv%\",";
      $record .= "\"Ki\",\" +/- \",\"cv%\",";
      $record .= "\"kinact/Ki\",\" +/- \",\"cv%\",";
   }
   print CSV $record, "\n";
   
   foreach (sort keys %main::resultsAver) {
      my $key = $_;
      my @av = split /\t/, $main::resultsAver{$key};
      my @sd = split /\t/, $main::resultsSdev{$key};
      print CSV "\"$key\",";
      my $n = scalar (@av);
      my $i = 0;
      for (; $i < $n; ++$i) {
         my $av = $av[$i];
         my $sd = $sd[$i];
         my $cv = 0;
         if ($av[$i] != 0) {
            $cv = sprintf ("%.1f", 100*abs($sd[$i]/$av[$i]));
         }
         print CSV "$av,$sd,$cv,";
      }
      print CSV "\n";
   }
}

#-------------------------------------------------------
sub write_report_htm
{
   my $record = "Cpd/Repl\tr.m.s.\t";
   $record .= "k<sub>s</sub><br>1/s\t +/- \tcv%\t";
   $record .= "k<sub>dI</sub><br>1/s\t +/- \tcv%\t";
   $record .= "k<sub>inact</sub><br>1/s\t +/- \tcv%\t";
   $record .= "K<sub>i</sub><br>&mu;M\t +/- \tcv%\t";
   $record .= "k<sub>inact</sub>/K<sub>i</sub><br>1/(&mu;M s)\t +/- \tcv%";
   
   $record =~ s|\t|\t<td class=\"th\">|g;

   print HTM "<h2>Individual replicates</h2>\n";
   print HTM "<table border=\"1\">\n<tr><td class=\"th\">$record\n";
   
   foreach (sort keys %main::resultsReplicates) {
      my $key = $_;
      my $record = $main::resultsReplicates{$key};
      my $rms = $main::rmsReplicates{$key};
      my @fields = split /\t/, $record;
      $key = "$fields[0]/$fields[1]";
      my $link = create_link ($key);
      print HTM "<tr><td>$link<td>$rms<td>";
      shift @fields;
      shift @fields;
      my $field;
      foreach $field (@fields) {
         $field = sprintf ("%g", $field);
      }
      print HTM join "<td>", @fields;
      print HTM "\n";
   }
   print HTM "</table>\n";
   if (!$main::ini{'UseTrueKi'}) {
      print_kiapp_warning_htm ();
   }

   $record = "Cpd\t";
   if ($main::ini{'AveragesGeometric'}) {
      $record .= "log(k<sub>s</sub>)<br>1/s\t +/- \tcv%\t";
      $record .= "log(k<sub>dI</sub>)<br>1/s\t +/- \tcv%\t";
      $record .= "log(k<sub>inact</sub>)<br>1/s\t +/- \tcv%\t";
      $record .= "log(K<sub>i</sub>)<br>&mu;M\t +/- \tcv%\t";
      $record .= "log(k<sub>inact</sub>/K<sub>i</sub>)<br>1/(&mu;M s)\t +/- \tcv%";
   }
   else {
      $record .= "k<sub>s</sub><br>1/s\t +/- \tcv%\t";
      $record .= "k<sub>dI</sub><br>1/s\t +/- \tcv%\t";
      $record .= "k<sub>inact</sub><br>1/s\t +/- \tcv%\t";
      $record .= "K<sub>i</sub><br>&mu;M\t +/- \tcv%\t";
      $record .= "k<sub>inact</sub>/K<sub>i</sub><br>1/(&mu;M s)\t +/- \tcv%";
   }
   
   $record =~ s|\t|\t<td class=\"th\">|g;

   print HTM "<h2>Averages and standard deviations</h2>\n";
   print HTM "<table border=\"1\">\n<tr><td class=\"th\">$record\n";
   
   foreach (sort keys %main::resultsAver) {
      my $key = $_;
      my @av = split /\t/, $main::resultsAver{$key};
      my @sd = split /\t/, $main::resultsSdev{$key};
      print HTM "<tr><td>$key\t";
      my $n = scalar (@av);
      my $i = 0;
      for (; $i < $n; ++$i) {
         my $av = $av[$i];
         my $sd = $sd[$i];
         my $cv = 0;
         if ($av[$i] != 0) {
            $cv = sprintf ("%.1f", 100*abs($sd[$i]/$av[$i]));
         }
         print HTM "<td>$av<td>\t$sd<td>\t$cv";
      }
      print HTM "\n";
   }
   print HTM "</table>\n";    
   if (!$main::ini{'UseTrueKi'}) {
      print_kiapp_warning_htm ();
   }
     
   my $linkTxt = "$main::ini{'ReportFile'}.txt";
   my $linkCsv = "$main::ini{'ReportFile'}.csv";
   print HTM <<_END;
 <p>
 <hr noshade size="3">
 <p>
 Text export: [<a href="$linkTxt">TXT</a>] [<a href="$linkCsv">CSV</a>]
_END
}

#-------------------------------------------------------
sub print_kiapp_warning_htm
{
   print HTM <<_END;
<p>
<p>
K<sub>i</sub> values listed are "true" K<sub>i</sub>'s, aassuming ATP-competitive mechanism.  
The "true" K<sub>i</sub> values were computed from "apparent" K<sub>i</sub>'s using the formula:
<p>
<p>
<center>
K<sub>i</sub><sup>(true)</sup> = K<sub>i</sub><sup>(app)</sup> / $main::ini{'KiAppAdjust'}
</center>
<p>
_END
}

#-------------------------------------------------------
sub write_parameters_file
{
   my $path = $main::ini{'ParametersFile'};
   open (OUT, ">$path") or error_exit ("could not open [$path] for writing!\n");
   
   my $record = "Cpd\t";
   $record .= "k(on)\t";
   $record .= "$main::ini{'kSubstrateLabel'}\t +/- \t";
   $record .= "kdI\t +/- \t";
   $record .= "kinact\t +/- \t";
   $record .= "Ki\t +/- \t";
   $record .= "kinact/Ki\t +/- \t";
   print OUT $record, "\n";
   
   foreach (sort keys %main::resultsAver) {
      my $key = $_;
      my @av = split /\t/, $main::resultsAver{$key};
      my @sd = split /\t/, $main::resultsSdev{$key};
      print OUT "$key\t";
      print OUT "$main::ini{'k(on)'}\t";
      my $n = scalar (@av);
      my $i = 0;
      
      for (; $i < $n; ++$i) {      
         my $av = $av[$i];
         my $sd = $sd[$i];
         if ($main::ini{'AveragesGeometric'}) {
            $av = exp($av*log(10.0));
            $sd = exp($sd*log(10.0));
         }
         $av = sprintf ("%g", $av);
         $sd = sprintf ("%g", $sd);
         print OUT "$av\t$sd\t";
      }
      print OUT "\n";
   }
   close (OUT);
}

#-------------------------------------------------------
sub create_link
{
   my ($key) = @_;
   my ($compound, $replicate) = split /\//, $key;
   my $prefix = "../";  # we have to back-track from the "results" directory!   This should be made programmatic.
   
   my $link = "<a href=\"$prefix$main::ini{'InputDirectory'}/$compound/$replicate/output/$main::ini{'ScriptFile'}/index.html\">";
   $link =~ s|\.txt||;
   $link .= "$key</a>";
   return $link;
}

#-------------------------------------------------------
sub process_files
{
   foreach (@main::files) {
      my $path = $_;
      process_script ($path);
   }
}

#-------------------------------------------------------
sub process_script
{
   my ($path) = @_;
   read_data ($path);
   my $script = write_script ($path);
   if ($main::ini{'ExecuteDynaFit'}) {
      execute_script ($script);
   }
   if ($main::ini{'CompileReport'}) {
      parse_output_params ($script);
      parse_output_summary ($script);
   }
}

#-------------------------------------------------------
sub write_script
{
   my ($path) = @_;
   $path =~ s|/[^/]+$||;
   $path =~ s|/data||;

   my $outputDir = $path;
   $outputDir =~ s|$main::ini{'InputDirectory'}|$main::ini{'ProjectDirectory'}|;
   $outputDir .= "/output/$main::ini{'ScriptFile'}";
   $outputDir =~ s|\.txt$||;

   my $dataDir = $path;
   $dataDir =~ s|$main::ini{'InputDirectory'}|$main::ini{'ProjectDirectory'}|;
   $dataDir .= "/data";

   my @fields = split /\//, $path;        
   my ($replicate) = pop @fields;
   my ($compound) = pop @fields;

   $path .= "/$main::ini{'ScriptFile'}";

   my $responseP = $main::ini{'Response'};
   my $kon = $main::ini{'k(on)'};
   my $Ki = $main::estimatedKi{$compound};
   if ($main::ini{'UseTrueKi'}) {
      $Ki /= $main::ini{'KiAppAdjust'};
   }
   my $koff = $kon * $Ki;
      
   my $responseBounds = $main::ini{'ResponseBounds'};
   my $enzConc = $main::ini{'EnzymeConcentration'};
   my $optEnz = "";
   my $kiAppAdjust = 1;
   if ($main::ini{'UseTrueKi'}) {
      $kiAppAdjust = $main::ini{'KiAppAdjust'};
   }
   my $KiApp = $Ki * $kiAppAdjust;
   if ($KiApp/$enzConc < $main::ini{'OptimEnzymeCutoff'}) {
      $optEnz = "?";
   }
   my $prodResp = $main::ini{'Response'};
   my $optResp = "";
   if ($main::ini{'ResponseOptimized'}) {
      $optResp = "?";
   }

   my @script = @main::scriptTemplate;
   my $line;
   foreach $line (@script) {

      $line =~ s|OUTPUTDIR|$outputDir|;
      $line =~ s|DATADIR|$dataDir|;
      $line =~ s|KON|$kon|;
      $line =~ s|KOFF|$koff|;
      $line =~ s|ENZCONC|$enzConc|;
      $line =~ s|OPTENZ|$optEnz|;
      $line =~ s|PRODRESP|$prodResp|;
      $line =~ s|OPTRESP|$optResp|;
      $line =~ s|RESPONSE_BOUNDS|$responseBounds|;

      if ($line =~ m|COLUMNS|) {
         $line = insert_columns ($compound, $replicate);
      }
   }
   
   open (OUT, ">$path") or error_exit ("could not open [$path]!\n");
   print OUT @script;
   close (OUT);
   
   if ($main::ini{'WriteBatchFiles'}) {
      my $pathBatch = $path;
      my $pathDynaFit = $path;
      $pathDynaFit =~ s|$main::ini{'InputDirectory'}/|$main::ini{'ProjectDirectory'}/|;
      
      my $backDir = $main::ini{'DynaFitDirectory'};
      $backDir =~ s|/|\\|g;
      my $batch = "cd $backDir\nDynaFitC $pathDynaFit\npause\n";

      $pathBatch =~ s|\.txt$|.bat|;
      open (OUT, ">$pathBatch") or error_exit ("could not open [$path]!\n");
      print OUT $batch;
      close (OUT);
   }

   return $path;
}

#-------------------------------------------------------
sub insert_columns
{
   my ($compound, $replicate) = @_;
   my $offsetBounds = $main::ini{'OffsetBounds'};
   my $columnsText = "";
   my $i;
   my $offsetValue = $main::ini{'LocalOffsetValue'};
   my $countUsedColumns = 0;
   my $key = "$compound/$replicate";
   my $record = $main::algebOffsets{$key};
   my @offsetAlgeb = split /\t/, $record;
   for ($i = 1; $i <= $main::ini{'Columns'}; ++$i) {
      if ($main::includeConc[$i-1] eq "+") {
         my $iCol = sprintf ("%3d", $i + 1);
         my $conc = $main::concs[$i-1];
         if ($conc == 0 && !$main::ini{'IncludeZero'}) {
            next;
         }
         $countUsedColumns++;
         my $opt = "";
         if ($conc > 0) {
            if ($main::ini{'OptimizeConcsAll'}) {
               if ($countUsedColumns != $main::ini{'HoldFixedConc'}) {
                  $opt = " ?";
               }
            }
            elsif ($countUsedColumns <= $main::ini{'OptimizeConcFirst'}) {
               $opt = " ?";
            }
         }
         my $offset = "";
         if ($main::ini{'OffsetFromAlgebraicFit'}) {
            $offsetValue = $offsetAlgeb[$countUsedColumns-1];
            if (!defined($offsetValue)) {
               $offsetValue = $main::ini{'LocalOffsetValue'};
            }
         }
         if ($main::ini{'LocalOffsetOptimize'}) {
            $offset = " offset = $offsetValue ? $offsetBounds |";
         }
         $columnsText .= "   column $iCol |$offset conc I = $conc$opt\n";
      }
   }
   return $columnsText;
}

#-------------------------------------------------------
sub parse_output_summary
{
   my ($script) = @_;
   my $outputDir = $script;
   $outputDir =~ s|$main::ini{'ScriptFile'}|output|;
   my $outputSubdir = $main::ini{'ScriptFile'};
   $outputSubdir =~ s|\.txt||;
   my $outputFile = "$outputDir/$outputSubdir/html/summary-t001.html";
   open (IN, $outputFile) or die "could not open [$outputFile]\n";
   my @txt = <IN>;
   close (IN);

   # r.m.s. deviation<TD>119.524

   my $txt = join "", @txt;
   $txt =~ m|r.m.s. deviation<TD>(.+)|;
   my $rms = $1;
   my $label = $script;
   $label =~ s|/$main::ini{'ScriptFile'}||;
   $label =~ s|$main::ini{'InputDirectory'}/||;
   $label =~ s|/R|\tR|;
   $main::rmsReplicates{$label} = $rms;
}

#-------------------------------------------------------
sub parse_output_params
{
   my ($script) = @_;
   my $outputDir = $script;
   $outputDir =~ s|$main::ini{'ScriptFile'}|output|;
   my $outputSubdir = $main::ini{'ScriptFile'};
   $outputSubdir =~ s|\.txt||;
   my $outputFile = "$outputDir/$outputSubdir/txt/parameters-t001-LM.txt";

   open (IN, $outputFile) or die "could not open [$outputFile]\n";
   my @txt = <IN>;
   close (IN);

   $script =~ s|/$main::ini{'ScriptFile'}||;
   $script =~ s|$main::ini{'InputDirectory'}/||;
   $script =~ s|/R|\tR|;
   my ($Ki, $KiEr, $kinact, $kinactEr);
   my $label = $script;
   my $record = "$label\t";
   line:
   foreach (@txt) {
      chomp ();
      my $line = $_;
      $line =~ s| ||g;
      my @fields = split /\t/, $line;
      my $n = scalar (@fields);
      if ($n < 4) {
         next line;
      }
      if ($fields[0] eq "NO.") {
         next line;
      }
      if ($fields[2] && $fields[2] eq $main::ini{'kSubstrateLabel'}) {
         my $cv = sprintf ("%.1f", 100*$fields[7]/$fields[6]);
         $record .= "$fields[6]\t$fields[7]\t$cv\t";
      }
      elsif ($fields[2] && $fields[2] eq 'kdI') {
         my $cv = sprintf ("%.1f", 100*$fields[7]/$fields[6]);
         $record .= "$fields[6]\t$fields[7]\t$cv\t";
         $Ki = $fields[6]/$main::ini{'k(on)'};
         $KiEr = $fields[7]/$main::ini{'k(on)'};
                  
         my $isTrueKi = $main::ini{'UseTrueKi'};
         if (!$isTrueKi) {
            my $div = $main::ini{'KiAppAdjust'};
            $Ki /= $div;
            $KiEr /= $div;
         }
      }
      elsif ($fields[2] && $fields[2] eq 'kinact') {
         my $cv = sprintf ("%.1f", 100*$fields[7]/$fields[6]);
         $kinact = $fields[6];
         $kinactEr = $fields[7];
         $record .= "$fields[6]\t$fields[7]\t$cv\t";
      }
   }
   if ($Ki && $Ki != 0.0) {
      my $cv = sprintf ("%.1f", 100*$KiEr/$Ki);
      $record .= "$Ki\t$KiEr\t$cv\t";
      my $kinactKi = $kinact/$Ki;
      my $kinactKiEr = $kinactKi * sqrt (($KiEr/$Ki)*($KiEr/$Ki) + ($kinactEr/$kinact)*($kinactEr/$kinact));
      $cv = sprintf ("%.1f", 100*$kinactKiEr/$kinactKi);
      $record .= "$kinactKi\t$kinactKiEr\t$cv\t";
   }   
   $main::resultsReplicates{$label} = $record;
}
 
#-------------------------------------------------------
sub execute_script
{
   my ($script) = @_;
   my $cwd = cwd ();
   $script =~ s|$main::ini{'InputDirectory'}|$main::ini{'ProjectDirectory'}|;
   my $log = $script;
   $log =~ s|\.txt$|.log|;

   chdir ($main::ini{'DynaFitDirectory'});
   my $command = "$main::ini{'DynaFitExe'} $script >>$log";
   my $screen = $script;
   $screen =~ s|$main::ini{'ProjectDirectory'}/||;
   $screen =~ s|/$main::ini{'ScriptFile'}||;
   $screen =~ s|/| : |;
   print $screen, " ... ";
   my $status = system ($command);
   print " status = $status\n";
   chdir ($cwd)
}

#-------------------------------------------------------
sub read_script_template
{
   my $path = $main::ini{'ScriptTemplate'};
   open (IN, $path) or error_exit ("could not open [$path]!\n");
   @main::scriptTemplate = <IN>;
   close (IN);
}

#-------------------------------------------------------
sub create_report_files
{
   my $path = $main::ini{'HtmlHeaderFile'};
   open (IN, $path) or error_exit ("could not open [$path]!\n");
   @main::htmlHeader = <IN>;
   close (IN);

   $path = $main::ini{'HtmlFooterFile'};
   open (IN, $path) or error_exit ("could not open [$path]!\n");
   @main::htmlFooter = <IN>;
   close (IN);

   my $status = mkdir ($main::ini{'ReportDirectory'});

   $path = "$main::ini{'ReportDirectory'}/$main::ini{'ReportFile'}.txt";
   open (TXT, ">$path") or error_exit ("could not open [$path] for writing!\n");
   
   $path = "$main::ini{'ReportDirectory'}/$main::ini{'ReportFile'}.csv";
   open (CSV, ">$path") or error_exit ("could not open [$path] for writing!\n");
   
   $path = "$main::ini{'ReportDirectory'}/$main::ini{'ReportFile'}.html";
   open (HTM, ">$path") or error_exit ("could not open [$path] for writing!\n");
   
   print HTM @main::htmlHeader; 
}

#-------------------------------------------------------
sub close_report_files
{
   print HTM @main::htmlFooter; 

   close (TXT);
   close (HTM);

   if (!$main::ini{'ExecuteDynaFit'} && !$main::ini{'CompileReport'}) {
      return;
   }

   my $path = "$main::ini{'ReportDirectory'}/$main::ini{'ReportFile'}.html";
   my $command = "start $path";
   system ($command);
}

#-------------------------------------------------------
sub filter_files
{
   my @files;
   foreach (@main::files) {
      my $path = $_;
      if ($path =~ m|$main::ini{'SheetFileName'}\.$main::ini{'FileExtension'}|) {
         push @files, $path;
      }
   }
   @main::files = @files;
}

#-------------------------------------------------------
sub conc_scale_factor 
{
   my $mult = 1;
   my $unit = $main::ini{'ConcentrationUnit'};
   if ($unit =~ m|pM|) {
      $mult = 1.e-12;
   }
   elsif ($unit =~ m|nM|) {
      $mult = 1.e-9;
   }
   elsif ($unit =~ m|uM|) {
      $mult = 1.e-6;
   }
   elsif ($unit =~ m|mM|) {
      $mult = 1.e-3;
   }
   elsif ($unit =~ m|M|) {
      $mult = 1;
   }
   else {
      print "ERROR: wrong concentration unit.\n",
            "Allowable units are M, mM, uM, nM, and pM.";
      termninate ();
   }
   return $mult;
}

#-------------------------------------------------------
sub read_data
{
   my ($path) = @_;

   open (IN, $path) or error_exit ("could not open [$path]!\n");
   my @txt = <IN>;
   close (IN);
      
   foreach (@txt) {
      if (/^\[I\]/) {
         chomp ();
         my @fields = split /\t/;         
         my $header = shift @fields;
         my ($label, $unit) = split ",", $header;
         my $multInput = 1;
         if ($unit =~ m|pM|) {
            $multInput = 1.e-12;
         }
         elsif ($unit =~ m|nM|) {
            $multInput = 1.e-9;
         }
         elsif ($unit =~ m|uM|) {
            $multInput = 1.e-6;
         }
         elsif ($unit =~ m|mM|) {
            $multInput = 1.e-3;
         }
         elsif ($unit =~ m|M|) {
            $multInput = 1;
         }
         else {
            print "ERROR: wrong unit for inhibitor concentration.\n",
                  "Allowable units are M, mM, uM, nM, and pM.";
            termninate ();
         }
         my $multConc = conc_scale_factor ();
         my $mult = $multInput/$multConc;        
         my $conc;
         foreach $conc (@fields) {
            $conc *= $mult;
         }
         @main::concs = @fields;
      }
      elsif (/INCLUDE/) {
         chomp ();
         my @fields = split /\t/;
         shift @fields;
         @main::includeConc = @fields;
      }
   }
}

#-------------------------------------------------------
#-------------------------------------------------------
sub initialize
{
	my ($line, $inifile, $logfile);

	if (@ARGV) {
		$inifile = $ARGV[0];
	}
	else {
		$inifile = "$main::program.ini";
	}
	read_inifile ($inifile);

	$logfile = "$main::program.log";
	open (LOG, ">$logfile") or die $!
}

#-------------------------------------------------------
sub read_inifile
{
	my ($inifile) = @_;
	my ($line, $lineRaw, $key, $value, $contin);

	open (INI, $inifile) or die "could not open initialization file [$inifile]!\n";

	$contin = 0;
	$line = "";

	lineRaw:
	foreach $lineRaw (<INI>) {
		chomp ($lineRaw);
		
		# Allow for continuation lines:

		if ($lineRaw =~ m|\\$|) {
			$contin = 1;
		   $lineRaw =~ s|\\$||;
			$line .= $lineRaw;
			next lineRaw;
		}
		else {
			if ($contin) {
				$line .= $lineRaw;
			}
			else {
				$line = $lineRaw;
			}
			$contin = 0;
		}

		$line =~ s|\;.+||g;
		if (!($line =~ m|=|)) {
			next;
		}

		($key, $value) = split /=/, $line;
		$key =~ s|^\s+||;
		$key =~ s|\s+$||;
		$value =~ s|^\s+||;
		$value =~ s|\s+$||;
		$main::ini{$key} = $value;

		$line = "";
	}
	close (INI);
}

#-------------------------------------------------
sub find_files_dir_ext
{
	my ($directory, $extension) = @_;

	undef @main::files;

	$main::match = "";
	if (length($extension) > 0) {
		$main::match = "\.$extension";
	}
	find (\&wanted_files, $directory);

	return @main::files;
}

#-------------------------------------------------
sub find_all_files
{
	my (@tmp);

	@tmp = find_file_extension ("$main::ini{'InputDirectory'}", "", "");

	return @tmp;
}

#-------------------------------------------------
sub find_files
{
	my (@tmp);

	@tmp = find_file_extension ("$main::ini{'InputDirectory'}", "", $main::ini{'FileExtension'});

	return @tmp;
}

#-------------------------------------------------
sub find_file_extension
{
	my ($directory, $file, $extension) = @_;

	undef @main::files;

	$main::match = "";
	if (defined($file) && length($file) > 0) {
		$main::match = $file;
	}
	elsif (length($extension) > 0) {
		$main::match = "\.$extension";
	}

	find (\&wanted_files, $directory);

	return @main::files;
}

#-------------------------------------------------
sub wanted_files
{
	m|$main::match$|i && push @main::files, $File::Find::name;
}

#-------------------------------------------------------
sub terminate
{
	close (LOG);
	if ((defined $main::ini{'PressEnterToFinish'}) && ($main::ini{'PressEnterToFinish'} eq 'n')) {
	   return;
	}
	print "All done.  Press ENTER to terminate.";
	<STDIN>;
}

#-------------------------------------------------------
sub error_exit
{
   my ($message) = @_;
   print "ERROR: $message\n\nPress [Enter] to terminate.";
   <STDIN>;
   exit ();   
}
