#!/usr/bin/perl

use Cwd;
use File::Find;
use File::Copy;
use File::Path qw(make_path);

use strict;
package main;

($main::program = $0) =~ s,.*[\\/],,;
$main::program =~ s|\.pl||;
$main::program =~ s|\.exe||;

$| = 1;  # turn off buffering of STDOUT output

initialize ();

my $inputDir = "$main::ini{'GnuPlotDirectory'}/$main::ini{'ProjectDirectory'}";
$inputDir =~ s|\.\.\.|\.\.|;
$main::ini{'InputDirectory'} = $inputDir;

run ();

#-------------------------------------------------------
sub run
{
   print <<_END;

============================================================
$main::ini{'Title'}
============================================================

_END
   if (-e "fit.log") {
      my $status = system ("del fit.log");
   }  
   create_report_files ();
   copy_template_files ();
   read_cellular_ic50 ();
   read_biochemical_params ();
   write_ic50_params_file ();
   log_linear_regression ();
   plot_params ();
   write_report ();
   close_report_files ();
   delete_temporary_files ();
}

#-------------------------------------------------------
sub delete_temporary_files
{
   my $command = "del fit.log";
   my $status = system ($command);

   $command = "del /q *.dat";
   $status = system ($command);
   
   $command = "del /q *.fit";
   $status = system ($command);
}

#-------------------------------------------------------
sub copy_template_files
{
   my $command = "copy ./templates/*.txt $main::ini{'ReportDirectory'}";
   $command =~ s|/|\\|g;
   my $status = system ($command);
}

#-------------------------------------------------------
sub write_ic50_params_file
{
   my $path = "$main::ini{'ReportDirectory'}/ic50params.txt";
   open (OUT, ">$path") or error_exit ("could not open [$path]!\n");
   my @compounds = sort keys %main::ic50;
   print OUT "Compound\tIc50\tkinact\tKi\tkinact/Ki\n";
   foreach (@compounds) {
      my $compound = $_;
      my $recordIc50 = $main::ic50{$compound};
      my $recordParams = $main::params{$compound};
      my $line = "$compound\t$recordIc50\t$recordParams";
      print OUT $line, "\n";
   }
   close (OUT);

   $path = "$main::ini{'ReportDirectory'}/ic50params-log.txt";
   open (OUT, ">$path") or error_exit ("could not open [$path]!\n");
   print OUT "Compound\tlog(Ic50)\tlog(kinact)\tlog(Ki)\tlog(kinact/Ki)\n";
   foreach (@compounds) {
      my $compound = $_;
      my $recordIc50 = $main::ic50Log{$compound};
      my $recordParams = $main::paramsLog{$compound};
      my $line = "$compound\t$recordIc50\t$recordParams";
      print OUT $line, "\n";
   }
   close (OUT);
}

#-------------------------------------------------------
sub plot_params
{
   my $cwd = cwd();
   chdir ($main::ini{'ReportDirectory'});
   
   my $path = "../$main::ini{'GnuPlotDirectory'}/$main::ini{'GnuPlotExe'}";
   $path =~ s|/|\\|g;
   
   my $commandGif = "$path gnuplot-gif.txt";
   my $statusGif = system ($commandGif);
   
   my $commandEps = "$path gnuplot-eps.txt";
   my $statusEps = system ($commandEps);
   
   chdir ($cwd);
}

#-------------------------------------------------------
sub fit_param
{
   my ($label, $data) = @_;
   
   my $path = "$label.dat";
   open (OUT, ">$path") or error_exit ("could not open [$path]!\n");
   print OUT $data;
   close (OUT);

   $path = "$label.fit";
   open (OUT, ">$path") or error_exit ("could not open [$path]!\n");
   print OUT <<_END;
f(x) = a*x + b
fit f(x) '$label.dat' using 1:2 via a,b
_END
   close (OUT);

   my $gnuPlotPath = "$main::ini{'GnuPlotDirectory'}/$main::ini{'GnuPlotExe'}";
   $gnuPlotPath =~ s|/|\\|g;

   my $command = "$gnuPlotPath $label.fit";
   my $status = system ($command);
}

#-------------------------------------------------------
sub process_fit_logfile
{
   my ($fitLog) = @_;
   my $ln10 = log(10.0);
   my ($label, $a, $b);
   foreach (@$fitLog) {
      chomp ();
      my $line = $_;
      if ($line =~ m|data read from '(.+?)\.dat'|) {
         $label = $1;         
      }
      elsif ($line =~ m|a               = (.+?)        \+\/\-|) {
         $a = $1;
      }
      elsif ($line =~ m|b               = (.+?)        \+\/\-|) {
         $b = $1;
         
         compute_r_squared ($label, $a, $b);
         
         $main::slopeLog{$label} = $a;
       # $main::intercepLog{$label} = $b;
         
         my $x1 = $main::ic50min;
         my $x1Log = log($x1)/$ln10;
         my $y1Log = $a*$x1Log + $b;
         my $y1 = exp($ln10*$y1Log);
         
         my $x2 = $main::ic50max;
         my $x2Log = log($x2)/$ln10;
         my $y2Log = $a*$x2Log + $b;
         my $y2 = exp($ln10*$y2Log);
         
         $main::y1{$label} = $y1;
         $main::y2{$label} = $y2;
      }
   }
}

#-------------------------------------------------------
sub compute_r_squared
{
   my ($label, $a, $b) = @_;  
   my $iColumn = 0;
   if ($label eq "kinact") {
      $iColumn = 0;
   }
   elsif ($label eq "Ki") {
      $iColumn = 1;
   }
   elsif ($label eq "kinactKi") {
      $iColumn = 2;
   }
   my @ic50 = compile_data_array (\%main::ic50Log, 0);
   my @param = compile_data_array (\%main::paramsLog, $iColumn);   
   my $ssqTot = compute_total_ssq (\@param);
   my $ssqRes = compute_residual_ssq (\@ic50, \@param, $a, $b);
   my $R2 = 1 - $ssqRes/$ssqTot;
   $R2 = sprintf ("%.3f", $R2);
   $main::R2{$label} = $R2;
}


#-------------------------------------------------------
sub compute_total_ssq
{
   my ($x) = @_;
   my @x = @$x;  
   my $n = scalar (@x);
   my $sum = 0;
   my $i = 0;
   for (; $i < $n; ++$i) {
      $sum += $x[$i];
   }
   my $aver = $sum/$n;
   $sum = 0;
   for ($i = 0; $i < $n; ++$i) {
      my $dif = $x[$i] - $aver;
      $sum += $dif * $dif;
   }
   return $sum;
}

#-------------------------------------------------------
sub compute_residual_ssq
{
   my ($x, $yData, $a, $b) = @_;
   my @x = @$x;  
   my @yData = @$yData;  
   my $n = scalar (@x);
   my $sum = 0;
   my $i = 0;
   for (; $i < $n; ++$i) {
      my $yModel = $a*$x[$i] + $b;
      my $dif = $yData[$i] - $yModel;
      $sum += $dif * $dif;
   }
   return $sum;
}

#-------------------------------------------------------
sub compile_data_array
{
   my ($data, $iField) = @_;
   my %data = %$data;
   my @compounds = sort keys %data;
   my @data;
   foreach (@compounds) {
      my $compound = $_;
      my @fields = split /\t/, $data{$compound};
      my $field = $fields[$iField];
      push @data, $field;
   }
   return (@data);
}

#-------------------------------------------------------
sub log_linear_regression
{
   my @compounds = sort keys %main::ic50Log;
   my $kinactData;
   my $KiData;
   my $kinactKiData;
   foreach (@compounds) {
      my $compound = $_;
      my $recordIc50 = $main::ic50Log{$compound};
      my $recordParams = $main::paramsLog{$compound};          
      my @fieldsIc50 = split /\t/, $recordIc50;
      my @fieldsParams = split /\t/, $recordParams;
      $kinactData    .= $fieldsIc50[0] . "\t" . $fieldsParams[0] . "\n";
      $KiData        .= $fieldsIc50[0] . "\t" . $fieldsParams[1] . "\n";
      $kinactKiData  .= $fieldsIc50[0] . "\t" . $fieldsParams[2] . "\n";
   }
   fit_param ("kinact", $kinactData);
   fit_param ("Ki", $KiData);
   fit_param ("kinactKi", $kinactKiData);

   my $path = "fit.log";
   open (IN, $path) or error_exit ("could not open [$path]!\n");
   my @fitLog = <IN>;
   close (IN);
   
   process_fit_logfile (\@fitLog);
   write_line_endpoints ();
}

#-------------------------------------------------------
sub write_line_endpoints
{
   my $path = "$main::ini{'ReportDirectory'}/fitted-lines.txt";
   open (OUT, ">$path") or error_exit ("could not open [$path]!\n");
   print OUT "Ic50\tkinact\tKi\tkinactKi\n";
   print OUT "$main::ic50min\t$main::y1{'kinact'}\t$main::y1{'Ki'}\t$main::y1{'kinactKi'}\n";
   print OUT "$main::ic50max\t$main::y2{'kinact'}\t$main::y2{'Ki'}\t$main::y2{'kinactKi'}\n";
   close (OUT);
}

#-------------------------------------------------------
sub read_biochemical_params
{
   my $path = $main::ini{'ParmetersFile'};
   my $ln10 = log(10.0);
   open (IN, $path) or error_exit ("could not open [$path]!");
   my @txt = <IN>;
   close (IN);   
   my $header = shift @txt; # Discard header line
   chomp ($header);
   my @fields = split /\t/, $header;
   my $kinactCol = -1;
   my $KiCol = -1;
   my $kinactKiCol = -1;
   my $iColumn = 0;
   for (; $iColumn < scalar (@fields); ++$iColumn) {
      if ($fields[$iColumn] eq $main::ini{'HeaderKinact'}) {
         $kinactCol = $iColumn;
      }
      elsif ($fields[$iColumn] eq $main::ini{'HeaderKi'}) {
         $KiCol = $iColumn;
      }
      elsif ($fields[$iColumn] eq $main::ini{'HeaderKinactKi'}) {
         $kinactKiCol = $iColumn;
      }
   }
   foreach (@txt) {
      chomp ();
      my $line = $_;
      my @fields = split /\t/, $line;
      my $compound      = $fields[0];
      my $kinact        = $fields[$kinactCol];
      my $kinactError   = $fields[$kinactCol + 1];
      my $Ki            = $fields[$KiCol];
      my $KiError       = $fields[$KiCol + 1];
      my $kinactKi      = $fields[$kinactKiCol];
      my $kinactKiError = $fields[$kinactKiCol + 1];
      my $record        = "$kinact\t$kinactError\t$Ki\t$KiError\t$kinactKi\t$kinactKiError";
      $main::params{$compound} = $record;

      my $kinactLog     = log($kinact)/$ln10;
      my $KiLog         = log($Ki)/$ln10;
      my $kinactKiLog   = log($kinactKi)/$ln10;
      my $recordLog     = "$kinactLog\t$KiLog\t$kinactKiLog";
      $main::paramsLog{$compound} = $recordLog;
   }
 }
 
#-------------------------------------------------------
sub read_cellular_ic50
{
   my $ln10 = log(10.0);
   my $path = $main::ini{'Ic50DataFile'};
   open (IN, $path) or error_exit ("could not open [$path]!");
   my @txt = <IN>;
   close (IN);   
   shift @txt; # Discard header line
   
   my $first = 1;
   foreach (@txt) {
      chomp ();
      my ($compound, $ic50, $error) = split /\t/, $_;
      if ($first) {
         $main::ic50min = $main::ic50max = $ic50;
         $first = 0;
      }
      else {
         $main::ic50min = $ic50 < $main::ic50min ? $ic50 : $main::ic50min;
         $main::ic50max = $ic50 > $main::ic50max ? $ic50 : $main::ic50max;
      }
      $main::ic50{$compound} = "$ic50\t$error";    
      my $ic50Log = log($ic50)/$ln10;
      $main::ic50Log{$compound} = $ic50Log;
   }
}

#-------------------------------------------------------
sub conc_scale_factor 
{
   my ($unit) = @_;
   my $mult = 1;
   if ($unit =~ m|pM|) {
      $mult = 1.e-12;
   }
   elsif ($unit =~ m|nM|) {
      $mult = 1.e-9;
   }
   elsif ($unit =~ m|uM|) {
      $mult = 1.e-6;
   }
   elsif ($unit =~ m|mM|) {
      $mult = 1.e-3;
   }
   elsif ($unit =~ m|M|) {
      $mult = 1;
   }
   else {
      print "ERROR: wrong concentration unit.\n",
            "Allowable units are M, mM, uM, nM, and pM.";
      termninate ();
   }
   return $mult;
}

#-------------------------------------------------------
sub write_report
{
   write_report_txt ();
   write_report_csv ();
   write_report_htm ();
}

#-------------------------------------------------------
sub write_report_txt
{
   my $path = "$main::ini{'ReportDirectory'}/$main::ini{'ReportFile'}.txt";
   open (OUT, ">$path") or error_exit ("could not open [$path]!\n");
   
   my @compounds = sort keys %main::ic50;
   print OUT "CELLULAR IC50 VS. BIOCHEMICAL PARAMETERS\n";
   print OUT "Compound\tIc50\t\tkinact\t\tKi\t\tkinact/Ki\t\n";
   foreach (@compounds) {
      my $compound = $_;
      my $recordIc50 = $main::ic50{$compound};
      my $recordParams = $main::params{$compound};
      my $line = "$compound\t$recordIc50\t$recordParams";
      print OUT $line, "\n";
   }
   print OUT "\n";
   print OUT "LOG(CELLULAR IC50) VS. LOG(BIOCHEMICAL PARAMETERS)\n";
   print OUT "Compound\tlog(Ic50)\tlog(kinact)\tlog(Ki)\tlog(kinact/Ki)\n";
   foreach (@compounds) {
      my $compound = $_;
      my $recordIc50 = $main::ic50Log{$compound};
      my $recordParams = $main::paramsLog{$compound};
      my $line = "$compound\t$recordIc50\t$recordParams";
      print OUT $line, "\n";
   }
   close (OUT);
}

#-------------------------------------------------------
sub write_report_csv
{
   my $path = "$main::ini{'ReportDirectory'}/$main::ini{'ReportFile'}.csv";
   open (OUT, ">$path") or error_exit ("could not open [$path]!\n");
   
   my @compounds = sort keys %main::ic50;
   print OUT "CELLULAR IC50 VS. BIOCHEMICAL PARAMETERS\n";
   my $line = "Compound\tIc50\t\tkinact\t\tKi\t\tkinact/Ki\t";
   $line =~ s|\t|,|g;
   print OUT $line, "\n";
   foreach (@compounds) {
      my $compound = $_;
      my $recordIc50 = $main::ic50{$compound};
      my $recordParams = $main::params{$compound};
      my $line = "$compound\t$recordIc50\t$recordParams";
      $line =~ s|\t|,|g;
      print OUT $line, "\n";
   }
   print OUT "\n";
   print OUT "LOG(CELLULAR IC50) VS. LOG(BIOCHEMICAL PARAMETERS)\n";
   $line = "Compound\tlog(Ic50)\tlog(kinact)\tlog(Ki)\tlog(kinact/Ki)";
   $line =~ s|\t|,|g;
   print OUT $line, "\n";
   foreach (@compounds) {
      my $compound = $_;
      my $recordIc50 = $main::ic50Log{$compound};
      my $recordParams = $main::paramsLog{$compound};
      my $line = "$compound\t$recordIc50\t$recordParams";
      $line =~ s|\t|,|g;
      print OUT $line, "\n";
   }
   close (OUT);
}

#-------------------------------------------------------
sub write_report_htm
{
   print HTM <<_END;
<h2>Correlations: Cellular vs. Biochemical Potency</h2>

<hr noshade size="4">
<h3>k<sub>inact</sub></h3>
Log-log slope: <b>$main::slopeLog{'kinact'}</b>
<br>
<font color="red">R<sup>2</sup> = <b>$main::R2{'kinact'}</b></font>
<br>
<img src="cell-ic50-kinact.gif" alt="">

<hr noshade size="4">
<h3>K<sub>i</sub></h3>
Log-log slope: <b>$main::slopeLog{'Ki'}</b>
<br>
<font color="red">R<sup>2</sup> = <b>$main::R2{'Ki'}</b></font>
<br>
<img src="cell-ic50-Ki.gif" alt="">

<hr noshade size="4">
<h3>k<sub>inact</sub> / K<sub>i</sub></h3>
Log-log slope: <b>$main::slopeLog{'kinactKi'}</b>
<br>
<font color="red">R<sup>2</sup> = <b>$main::R2{'kinactKi'}</b></font>
<br>
<img src="cell-ic50-kinactKi.gif" alt="">
_END

   my $linkTxt = "$main::ini{'ReportFile'}.txt";
   my $linkCsv = "$main::ini{'ReportFile'}.csv";
   print HTM <<_END;
 <p>
 <hr noshade size="3">
 <p>
 Text export: [<a href="$linkTxt">TXT</a>] [<a href="$linkCsv">CSV</a>]
_END

}

#-------------------------------------------------------
sub create_report_files
{
   my $path = $main::ini{'HtmlHeaderFile'};
   open (IN, $path) or error_exit ("could not open file\n[$path]\n");
   @main::htmlHeader = <IN>;
   close (IN);

   $path = $main::ini{'HtmlFooterFile'};
   open (IN, $path) or error_exit ("could not open file\n[$path]\n");
   @main::htmlFooter = <IN>;
   close (IN);

   my $status = mkdir ($main::ini{'ReportDirectory'});
   
   $path = "$main::ini{'ReportDirectory'}/$main::ini{'ReportFile'}.txt";
   open (TXT, ">$path") or error_exit ("could not open output file\n[$path]\n");
   
   $path = "$main::ini{'ReportDirectory'}/$main::ini{'ReportFile'}.csv";
   open (CSV, ">$path") or error_exit ("could not open output file\n[$path]\n");
   
   $path = "$main::ini{'ReportDirectory'}/$main::ini{'ReportFile'}.html";
   open (HTM, ">$path") or error_exit ("could not open output file\n[$path]\n");
   
   print HTM @main::htmlHeader; 
}

#-------------------------------------------------------
sub close_report_files
{
   print HTM @main::htmlFooter; 

   close (TXT);
   close (CSV);
   close (HTM);

   my $path = "$main::ini{'ReportDirectory'}/$main::ini{'ReportFile'}.html";
   my $command = "start $path";
   system ($command);
}

#-------------------------------------------------------
#-------------------------------------------------------
sub initialize
{
	my ($line, $inifile, $logfile);

	if (@ARGV) {
		$inifile = $ARGV[0];
	}
	else {
		$inifile = "$main::program.ini";
	}
	read_inifile ($inifile);

	$logfile = "$main::program.log";
	open (LOG, ">$logfile") or error_exit ("could not open log file\n[$logfile]\n");
}

#-------------------------------------------------------
sub read_inifile
{
	my ($inifile) = @_;
	my ($line, $lineRaw, $key, $value, $contin);

	open (INI, $inifile) or error_exit ("could not open initialization file\n[$inifile]\n");

	$contin = 0;
	$line = "";

	lineRaw:
	foreach $lineRaw (<INI>) {
		chomp ($lineRaw);
		
		# Allow for continuation lines:

		if ($lineRaw =~ m|\\$|) {
			$contin = 1;
		   $lineRaw =~ s|\\$||;
			$line .= $lineRaw;
			next lineRaw;
		}
		else {
			if ($contin) {
				$line .= $lineRaw;
			}
			else {
				$line = $lineRaw;
			}
			$contin = 0;
		}

		$line =~ s|\;.+||g;
		if (!($line =~ m|=|)) {
			next;
		}

		($key, $value) = split /=/, $line;
		$key =~ s|^\s+||;
		$key =~ s|\s+$||;
		$value =~ s|^\s+||;
		$value =~ s|\s+$||;
		$main::ini{$key} = $value;

		$line = "";
	}
	close (INI);
}

#-------------------------------------------------
sub find_files_dir_ext
{
	my ($directory, $extension) = @_;

	undef @main::files;

	$main::match = "";
	if (length($extension) > 0) {
		$main::match = "\.$extension";
	}
	find (\&wanted_files, $directory);

	return @main::files;
}

#-------------------------------------------------
sub find_all_files
{
	my (@tmp);

	@tmp = find_file_extension ("$main::ini{'InputDirectory'}", "", "");

	return @tmp;
}

#-------------------------------------------------
sub find_files
{
	my (@tmp);

	@tmp = find_file_extension ("$main::ini{'InputDirectory'}", "", $main::ini{'FileExtension'});

	return @tmp;
}

#-------------------------------------------------
sub find_file_extension
{
	my ($directory, $file, $extension) = @_;

	undef @main::files;

	$main::match = "";
	if (defined($file) && length($file) > 0) {
		$main::match = $file;
	}
	elsif (length($extension) > 0) {
		$main::match = "\.$extension";
	}

	find (\&wanted_files, $directory);

	return @main::files;
}

#-------------------------------------------------
sub wanted_files
{
	m|$main::match$|i && push @main::files, $File::Find::name;
}

#-------------------------------------------------------
sub terminate
{
	close (LOG);
	if ((defined $main::ini{'PressEnterToFinish'}) && ($main::ini{'PressEnterToFinish'} eq 'n')) {
	   return;
	}
	print "All done.  Press ENTER to terminate.";
	<STDIN>;
}

#-------------------------------------------------------
sub error_exit
{
   my ($message) = @_;
   print "ERROR: $message\n\nPress [Enter] to terminate.";
   <STDIN>;
   exit ();   
}
